"""
Usage:
    local_runner.py --raw-data=<files> --streaming-bts=<file> [--window-bts=<file>] [--output-file=<file>]
    local_runner.py (-h | --help)
"""
import csv
import json
from collections import defaultdict
from typing import List, Optional, Dict, Tuple

from smart_open import smart_open

from blurr.cli.validate import validate
from blurr.runner.data_processor import DataProcessor, SimpleJsonDataProcessor
from blurr.runner.json_encoder import BlurrJSONEncoder
from blurr.runner.runner import Runner, TimeAndRecord


class LocalRunner(Runner):
    def __init__(self, stream_bts_file: str, window_bts_file: Optional[str] = None):
        super().__init__(stream_bts_file, window_bts_file)

        self._per_user_data = {}

    def _validate_bts_syntax(self) -> None:
        validate(self._stream_bts)
        if self._window_bts is not None:
            validate(self._window_bts)

    def _execute_for_all_identities(self,
                                    identity_records: Dict[str, List[TimeAndRecord]],
                                    old_state: Optional[Dict[str, Dict]] = None) -> None:
        if not old_state:
            old_state = {}
        for identity, records in identity_records.items():
            _, data = self.execute_per_identity_records(identity, records,
                                                        old_state.get(identity, None))
            self._per_user_data[identity] = data

        for identity, state in old_state.items():
            if identity not in self._per_user_data:
                self._per_user_data[identity] = (old_state[identity], [])

    def get_identity_records_from_json_files(
            self, json_files: List[str], data_processor: DataProcessor = SimpleJsonDataProcessor()
    ) -> Dict[str, List[TimeAndRecord]]:
        identity_records = defaultdict(list)
        for file in json_files:
            with smart_open(file) as file_stream:
                for identity, record_with_datetime in self.get_per_identity_records(
                        file_stream, data_processor):
                    identity_records[identity].append(record_with_datetime)
        return identity_records

    def execute(self,
                identity_records: Dict[str, List[TimeAndRecord]],
                old_state: Optional[Dict[str, Dict]] = None) -> Dict[str, Tuple[Dict, List]]:
        self._execute_for_all_identities(identity_records, old_state)
        return self._per_user_data

    def print_output(self, per_user_data) -> None:
        for id, (block_data, window_data) in per_user_data.items():
            if not self._window_bts:
                for data in block_data.items():
                    print(json.dumps(data, cls=BlurrJSONEncoder))
            else:
                print(json.dumps((id, window_data), cls=BlurrJSONEncoder))

    def write_output_file(self, output_file: str, per_user_data):
        if not self._window_bts:
            with smart_open(output_file, 'w') as file:
                for block_data, _ in per_user_data.values():
                    for row in block_data.items():
                        file.write(json.dumps(row, cls=BlurrJSONEncoder))
                        file.write('\n')
        else:
            header = []
            for _, window_data in per_user_data.values():
                for data_row in window_data:
                    header = list(data_row.keys())
                    break
            header.sort()
            with smart_open(output_file, 'w') as csv_file:
                writer = csv.DictWriter(csv_file, header)
                writer.writeheader()
                for _, window_data in per_user_data.values():
                    writer.writerows(window_data)



from datetime import datetime
from typing import Dict

from blurr.core.aggregate import AggregateSchema, Aggregate
from blurr.core.evaluation import EvaluationContext
from blurr.core.field import FieldSchema, Field
from blurr.core.loader import TypeLoader
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store_key import Key, KeyType
from blurr.core.type import Type
from blurr.core.validator import validate_enum_attribute


class IdentityAggregateSchema(AggregateSchema):
    """
    Aggregates that handles the block rollup aggregation
    """
    ATTRIBUTE_DIMENSIONS = 'Dimensions'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.dimension_fields: Dict[str, FieldSchema] = {
            schema_spec[self.ATTRIBUTE_NAME]: self.schema_loader.get_nested_schema_object(
                self.fully_qualified_name, schema_spec[self.ATTRIBUTE_NAME])
            for schema_spec in self._spec.get(self.ATTRIBUTE_DIMENSIONS, [])
        }
        self.key_type = KeyType.DIMENSION

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_STORE)
        for schema_spec in self._spec.get(self.ATTRIBUTE_DIMENSIONS, []):
            self.add_errors(
                validate_enum_attribute(
                    self.schema_loader.get_fully_qualified_name(self.fully_qualified_name,
                                                                schema_spec[self.ATTRIBUTE_NAME]),
                    schema_spec, self.ATTRIBUTE_TYPE, {Type.INTEGER.value, Type.STRING.value}))


class IdentityAggregate(Aggregate):
    def __init__(self, schema: AggregateSchema, identity: str,
                 evaluation_context: EvaluationContext) -> None:
        super().__init__(schema, identity, evaluation_context)

        self._dimension_fields: Dict[str, Field] = {
            name: TypeLoader.load_item(item_schema.type)(item_schema, self._evaluation_context)
            for name, item_schema in self._schema.dimension_fields.items()
        }

        # Also add the dimension fields to regular fields so that they get processed by other
        # functions such as snapshot, restore, etc as per normal.
        # We don't add self._dimension_fields here as we want these fields to be separate objects.
        self._fields.update({
            name: TypeLoader.load_item(item_schema.type)(item_schema, self._evaluation_context)
            for name, item_schema in self._schema.dimension_fields.items()
        })
        self._existing_key = None

    def run_evaluate(self) -> None:
        if not self._needs_evaluation:
            return

        if not self._evaluate_dimension_fields():
            return

        # First time being run. Load state from store.
        if not self._existing_key:
            self._init_state_from_new_key()

        if self._existing_key and not self._compare_dimensions_to_fields():
            self._persist()
            self._init_state_from_new_key()

        super().run_evaluate()
        self._existing_key = self._key

    def _init_state_from_new_key(self):
        snapshot = self._store.get(self._key)
        if snapshot:
            self.run_restore(snapshot)
        else:
            self.run_reset()

    def _evaluate_dimension_fields(self) -> bool:
        """
        Evaluates the dimension fields. Returns False if any of the fields could not be evaluated.
        """
        for _, item in self._dimension_fields.items():
            item.run_evaluate()
            if item.eval_error:
                return False
        return True

    def _compare_dimensions_to_fields(self) -> bool:
        """ Compares the dimension field values to the value in regular fields."""
        for name, item in self._dimension_fields.items():
            if item.value != self._nested_items[name].value:
                return False
        return True

    @property
    def _key(self):
        """ Generates the Key object based on dimension fields. """
        return Key(self._schema.key_type, self._identity, self._name,
                   [str(item.value) for item in self._dimension_fields.values()])

    def _persist(self) -> None:
        if self._existing_key:
            self._store.save(self._existing_key, self._snapshot)



from typing import Dict, Any, Type, Optional, Union, List, Tuple, Callable, Set

from blurr.core.errors import InvalidIdentifierError, RequiredAttributeError, EmptyAttributeError, \
    InvalidNumberError, InvalidValueError

ATTRIBUTE_NAME = 'Name'
ATTRIBUTE_TYPE = 'Type'
ATTRIBUTE_INTERNAL = '_Internal'


def validate_python_identifier_attributes(fully_qualified_name: str, spec: Dict[str, Any],
                                          *attributes: str) -> List[InvalidIdentifierError]:
    """ Validates a set of attributes as identifiers in a spec """
    errors: List[InvalidIdentifierError] = []

    checks: List[Tuple[Callable, InvalidIdentifierError.Reason]] = [
        (lambda x: x.startswith('_'), InvalidIdentifierError.Reason.STARTS_WITH_UNDERSCORE),
        (lambda x: x.startswith('run_'), InvalidIdentifierError.Reason.STARTS_WITH_RUN),
        (lambda x: not x.isidentifier(), InvalidIdentifierError.Reason.INVALID_PYTHON_IDENTIFIER),
    ]

    for attribute in attributes:
        if attribute not in spec or spec.get(ATTRIBUTE_INTERNAL, False):
            continue

        for check in checks:
            if check[0](spec[attribute]):
                errors.append(
                    InvalidIdentifierError(fully_qualified_name, spec, attribute, check[1]))
                break

    return errors


def validate_required_attributes(fully_qualified_name: str, spec: Dict[str, Any],
                                 *attributes: str) -> List[RequiredAttributeError]:
    """ Validates to ensure that a set of attributes are present in spec """
    return [
        RequiredAttributeError(fully_qualified_name, spec, attribute)
        for attribute in attributes
        if attribute not in spec
    ]


def validate_empty_attributes(fully_qualified_name: str, spec: Dict[str, Any],
                              *attributes: str) -> List[EmptyAttributeError]:
    """ Validates to ensure that a set of attributes do not contain empty values """
    return [
        EmptyAttributeError(fully_qualified_name, spec, attribute)
        for attribute in attributes
        if not spec.get(attribute, None)
    ]


def validate_number_attribute(
        fully_qualified_name: str,
        spec: Dict[str, Any],
        attribute: str,
        value_type: Union[Type[int], Type[float]] = int,
        minimum: Optional[Union[int, float]] = None,
        maximum: Optional[Union[int, float]] = None) -> Optional[InvalidNumberError]:
    """ Validates to ensure that the value is a number of the specified type, and lies with the specified range """

    if attribute not in spec:
        return

    try:
        value = value_type(spec[attribute])
        if (minimum is not None and value < minimum) or (maximum is not None and value > maximum):
            raise None
    except:
        return InvalidNumberError(fully_qualified_name, spec, attribute, value_type, minimum,
                                  maximum)


def validate_enum_attribute(fully_qualified_name: str, spec: Dict[str, Any], attribute: str,
                            candidates: Set[Union[str, int, float]]) -> Optional[InvalidValueError]:
    """ Validates to ensure that the value of an attribute lies within an allowed set of candidates """

    if attribute not in spec:
        return

    if spec[attribute] not in candidates:
        return InvalidValueError(fully_qualified_name, spec, attribute, candidates)



from decimal import Decimal
from json import JSONEncoder

from blurr.core.store_key import Key


class BlurrJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            ratio = o.as_integer_ratio()
            if ratio[1] == 1:
                return int(o)
            else:
                return float(o)

        if isinstance(o, Key):
            return str(o)

        return super().default(o)



from abc import ABC, abstractmethod
from typing import Any

from blurr.core import logging
from blurr.core.base import BaseSchema, BaseItem
from blurr.core.evaluation import Expression, EvaluationContext
from blurr.core.schema_loader import SchemaLoader


class FieldSchema(BaseSchema, ABC):
    """
    An individual field schema.
        1. Field Schema must be defined inside a Group
        2. Contain the attributes:
            a. Name (inherited from base)
            b. Type (inherited from base)
            c. Value (required)
            d. Filter (inherited from base)
    """

    # Field Name Definitions
    ATTRIBUTE_VALUE = 'Value'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.value: Expression = self.build_expression(self.ATTRIBUTE_VALUE)

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_VALUE)

    @property
    @abstractmethod
    def type_object(self) -> Any:
        """
        Returns the type object the Type represents
        """
        raise NotImplementedError('type_object is required')

    @staticmethod
    def sanitize_object(instance: Any) -> Any:
        """
        Sanitize object after evaluation if needed
        """
        return instance

    def is_type_of(self, instance: Any) -> bool:
        """
        Checks if instance is of the type
        :param instance: An object instance
        :return: True if the object is of this type, False otherwise
        """
        return isinstance(instance, self.type_object)

    @property
    @abstractmethod
    def default(self) -> Any:
        """
        Returns the default value for this type
        """
        raise NotImplementedError('type_object is required')

    @staticmethod
    def encoder(value: Any) -> Any:
        return value

    # TODO: Add unit tests for the casting being done here.
    def decoder(self, value: Any) -> Any:
        return self.sanitize_object(self.type_object(value))


class Field(BaseItem):
    """
    An individual field object responsible for retaining the field value
    """

    def __init__(self, schema: FieldSchema, evaluation_context: EvaluationContext) -> None:
        """
        Initializes the Field with the default for the schema
        :param schema: Field schema
        :param evaluation_context: Context dictionary for evaluation
        """
        super().__init__(schema, evaluation_context)

        # When the field is created, the value is set to the field type default
        self.value = self._schema.default
        self.eval_error = False

    def run_evaluate(self) -> None:
        """
        Overrides the base evaluation to set the value to the evaluation result of the value
        expression in the schema
        """
        result = None
        self.eval_error = False
        if self._needs_evaluation:
            result = self._schema.value.evaluate(self._evaluation_context)

        self.eval_error = result is None
        if self.eval_error:
            return

        # Only set the value if it conforms to the field type
        if not self._schema.is_type_of(result):
            try:
                result = self._schema.type_object(result)
            except Exception as err:
                logging.debug('{} in casting {} to {} for field {}. Error: {}'.format(
                    type(err).__name__, result, self._schema.type,
                    self._schema.fully_qualified_name, err))
                self.eval_error = True
                return

        try:
            result = self._schema.sanitize_object(result)
        except Exception as err:
            logging.debug('{} in sanitizing {} of type {} for field {}. Error: {}'.format(
                type(err).__name__, result, self._schema.type, self._schema.fully_qualified_name,
                err))
            self.eval_error = True
            return

        self.value = result

    @property
    def _snapshot(self) -> Any:
        """
        Snapshots the current value of the field
        """
        return self._schema.encoder(self.value)

    def run_restore(self, snapshot) -> None:
        """
        Restores the value of a field from a snapshot
        """
        self.value = self._schema.decoder(snapshot)

    def run_reset(self) -> None:
        self.value = self._schema.default

    def __repr__(self):
        return str(self._snapshot)


class ComplexTypeBase(ABC):
    """
    Implements a wrapper for base methods declared in base types such that the current object
    is returned in cases there are no returned values.  This ensures that evaluating the `Value`
    expression for field always returns an object.
    """

    def __getattribute__(self, item):
        """ Overrides the default getattribute to return self"""

        # Resolve the attribute by inspecting the current object
        attribute = super().__getattribute__(item)

        # Return the attribute as-is when it is NOT a function
        if not callable(attribute) or (item.startswith('__') and item.endswith('__')):
            return attribute

        # Wrap the attribute in a function that changes its return value
        def wrapped_attribute(*args, **kwargs):

            # Executing the underlying method
            result = attribute(*args, **kwargs)

            # If the execution does not return a value
            if result is None:
                return self

            # Get the type of the current object
            self_type = type(self)

            # If the method executed is defined in the base type and a base type object is returned
            # (and not the current type), then wrap the base object into an object of the current type
            if isinstance(result, self_type.__bases__) and not isinstance(result, self_type):
                return self_type(result)
                # TODO This creates a shallow copy of the object - find a way to optimize this

            # Return the result as-is on all other conditions
            return result

        # Return the wrapped attribute instead of the default
        return wrapped_attribute



from typing import List, Optional

from blurr.cli.util import get_stream_window_bts_files, get_yml_files, eprint
from blurr.cli.validate import get_valid_yml_files
from blurr.runner.data_processor import IpfixDataProcessor, SimpleJsonDataProcessor, DataProcessor
from blurr.runner.local_runner import LocalRunner
from blurr.runner.spark_runner import SparkRunner

RUNNER_CLASS = {
    'local': LocalRunner,
    'spark': SparkRunner,
}

DATA_PROCESSOR_CLASS = {'ipfix': IpfixDataProcessor, 'simple': SimpleJsonDataProcessor}


def transform(runner: Optional[str], stream_bts_file: Optional[str], window_bts_file: Optional[str],
              data_processor: Optional[str], raw_json_files: List[str]) -> int:
    if stream_bts_file is None and window_bts_file is None:
        stream_bts_file, window_bts_file = get_stream_window_bts_files(
            get_valid_yml_files(get_yml_files()))

    if stream_bts_file is None:
        eprint('Streaming BTS not provided and could not be found in the current directory.')
        return 1

    if not runner:
        runner = 'local'

    if not data_processor:
        data_processor = 'simple'

    if runner not in RUNNER_CLASS:
        eprint('Unknown runner: \'{}\'. Possible values: {}'.format(runner,
                                                                    list(RUNNER_CLASS.keys())))
        return 1

    if data_processor not in DATA_PROCESSOR_CLASS:
        eprint('Unknown data-processor: \'{}\'. Possible values: {}'.format(
            runner, list(DATA_PROCESSOR_CLASS.keys())))
        return 1

    data_processor_obj = DATA_PROCESSOR_CLASS[data_processor]()
    if runner == 'local':
        return transform_local(stream_bts_file, window_bts_file, raw_json_files, data_processor_obj)
    else:
        return transform_spark(stream_bts_file, window_bts_file, raw_json_files, data_processor_obj)


def transform_spark(stream_bts_file: Optional[str], window_bts_file: Optional[str],
                    raw_json_files: List[str], data_processor: DataProcessor) -> int:
    runner = SparkRunner(stream_bts_file, window_bts_file)
    out = runner.execute(runner.get_record_rdd_from_json_files(raw_json_files, data_processor))
    runner.print_output(out)

    return 0


def transform_local(stream_bts_file: Optional[str], window_bts_file: Optional[str],
                    raw_json_files: List[str], data_processor: DataProcessor) -> int:
    runner = LocalRunner(stream_bts_file, window_bts_file)
    out = runner.execute(
        runner.get_identity_records_from_json_files(raw_json_files, data_processor))
    runner.print_output(out)

    return 0



from datetime import datetime
from typing import Any, Dict, List, Tuple

import boto3
from boto3.dynamodb.conditions import Key as DynamoKey, Attr
from dateutil import parser

from blurr.core.schema_loader import SchemaLoader
from blurr.core.store import Store, Key, StoreSchema
from blurr.core.store_key import KeyType


class DynamoStoreSchema(StoreSchema):
    ATTRIBUTE_TABLE = 'Table'
    ATTRIBUTE_READ_CAPACITY_UNITS = 'ReadCapacityUnits'
    ATTRIBUTE_WRITE_CAPACITY_UNITS = 'WriteCapacityUnits'
    QUERY_LIMIT = 1000

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)
        self.table_name = self._spec.get(self.ATTRIBUTE_TABLE, None)
        self.rcu = self._spec.get(self.ATTRIBUTE_READ_CAPACITY_UNITS, 5)
        self.wcu = self._spec.get(self.ATTRIBUTE_WRITE_CAPACITY_UNITS, 5)
        self.query_limit = self.QUERY_LIMIT

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_TABLE)


class DynamoStore(Store):
    """
    Dynamo store implementation
    """

    def __init__(self, schema: DynamoStoreSchema) -> None:
        self._schema = schema
        self._dynamodb_resource = DynamoStore.get_dynamodb_resource()
        self._table = self._dynamodb_resource.Table(self._schema.table_name)

        # Test that the table exists.  Create a new one otherwise
        try:
            self._table.creation_date_time
        except self._dynamodb_resource.meta.client.exceptions.ResourceNotFoundException:
            self._table = self._dynamodb_resource.create_table(
                TableName=self._schema.table_name,
                KeySchema=[
                    {
                        'AttributeName': 'partition_key',
                        'KeyType': 'HASH'
                    },
                    {
                        'AttributeName': 'range_key',
                        'KeyType': 'RANGE'
                    },
                ],
                AttributeDefinitions=[{
                    'AttributeName': 'partition_key',
                    'AttributeType': 'S'
                }, {
                    'AttributeName': 'range_key',
                    'AttributeType': 'S'
                }],
                ProvisionedThroughput={
                    'ReadCapacityUnits': self._schema.rcu,
                    'WriteCapacityUnits': self._schema.wcu
                })
            # Wait until the table creation is complete
            self._table.meta.client.get_waiter('table_exists').wait(
                TableName=self._schema.table_name, WaiterConfig={'Delay': 5})

    @staticmethod
    # This is separate out as a separate function so that this can be mocked in unit tests.
    def get_dynamodb_resource() -> Any:
        return boto3.resource('dynamodb')

    @staticmethod
    def clean_for_get(item: Dict[str, Any]) -> Dict[str, Any]:
        item.pop('partition_key', None)
        item.pop('range_key', None)
        return item

    @staticmethod
    def clean_item_for_save(item: Dict[str, Any]) -> Dict[str, Any]:
        return {k: v for k, v in item.items() if v}

    def prepare_record(self, record: Dict[str, Any]) -> Tuple[Key, Any]:
        key = Key.parse_sort_key(record['partition_key'], record['range_key'])
        return key, self.clean_for_get(record)

    def get(self, key: Key) -> Any:
        item = self._table.get_item(Key={
            'partition_key': key.identity,
            'range_key': key.sort_key
        }).get('Item', None)

        if not item:
            return None

        return self.clean_for_get(item)

    def _get_range_timestamp_key(self, start: Key, end: Key,
                                 count: int = 0) -> List[Tuple[Key, Any]]:
        sort_key_condition = DynamoKey('range_key').between(start.sort_key, end.sort_key)
        # Limit is set to count+1 because for items where the start key matches exactly
        # KeyConditionExpression passes and FilterExpression fails.
        response = self._table.query(
            Limit=abs(count) + 1 if count else self._schema.query_limit,
            KeyConditionExpression=DynamoKey('partition_key').eq(
                start.identity) & sort_key_condition,
            FilterExpression=Attr('_start_time').gt(start.timestamp.isoformat()) &
            Attr('_start_time').lt(end.timestamp.isoformat()),
            ScanIndexForward=count >= 0,
        )
        items = sorted([self.prepare_record(item) for item in response.get('Items', [])])
        if count:
            items = self._restrict_items_to_count(items, count)
        return items

    def _get_range_dimension_key(self,
                                 base_key: Key,
                                 start_time: datetime,
                                 end_time: datetime = None,
                                 count: int = 0) -> List[Tuple[Key, Any]]:
        # A smaller limit cannot be set when abs(count) > 0 because all items need to be
        # returned to find the count number of elements in a sorted manner.
        # TODO: Improve count query performance by using a secondary index.
        response = self._table.query(
            Limit=self._schema.query_limit,
            KeyConditionExpression=DynamoKey('partition_key').eq(
                base_key.identity) & DynamoKey('range_key').begins_with(base_key.sort_prefix_key),
            FilterExpression=Attr('_start_time').gt(start_time.isoformat()) &
            Attr('_start_time').lt(end_time.isoformat()),
            ScanIndexForward=count >= 0,
        )
        items = sorted(
            [self.prepare_record(item) for item in response.get('Items', [])],
            key=lambda i: i[1].get('_start_time', datetime.min.isoformat()))
        if count:
            items = self._restrict_items_to_count(items, count)
        return items

    def get_all(self, identity: str) -> Dict[Key, Any]:
        response = self._table.query(KeyConditionExpression=DynamoKey('partition_key').eq(identity))
        return dict([self.prepare_record(item)
                     for item in response['Items']] if 'Items' in response else [])

    def save(self, key: Key, item: Any) -> None:
        item['partition_key'] = key.identity
        item['range_key'] = key.sort_key
        self._table.put_item(Item=self.clean_item_for_save(item))

    def delete(self, key: Key) -> None:
        pass

    def finalize(self) -> None:
        pass



from abc import ABC
from copy import copy
from typing import Dict

from blurr.core.aggregate import Aggregate
from blurr.core.base import BaseItemCollection, BaseSchemaCollection
from blurr.core.errors import MissingAttributeError
from blurr.core.loader import TypeLoader
from blurr.core.schema_context import SchemaContext
from blurr.core.schema_loader import SchemaLoader


class TransformerSchema(BaseSchemaCollection, ABC):
    """
    All Transformer Schema inherit from this base.  Adds support for handling
    the required attributes of a schema.
    """

    ATTRIBUTE_VERSION = 'Version'
    ATTRIBUTE_AGGREGATES = 'Aggregates'
    ATTRIBUTE_IMPORT = 'Import'
    ATTRIBUTE_STORES = 'Stores'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader, self.ATTRIBUTE_AGGREGATES)

        self.version = self._spec.get(self.ATTRIBUTE_VERSION, None)
        self.import_list = self._spec.get(self.ATTRIBUTE_IMPORT, [])
        self.schema_context = SchemaContext(self.import_list)

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_VERSION)
        self.validate_enum_attribute(self.ATTRIBUTE_VERSION, {'2018-03-01'})


class Transformer(BaseItemCollection, ABC):
    """
    All transformers inherit from this base.  Adds the current transformer
    to the context
    """

    def __init__(self, schema: TransformerSchema, identity: str) -> None:
        super().__init__(schema, copy(schema.schema_context.context))
        # Load the nested items into the item
        self._aggregates: Dict[str, Aggregate] = {
            name: TypeLoader.load_item(item_schema.type)(item_schema, identity,
                                                         self._evaluation_context)
            for name, item_schema in schema.nested_schema.items()
        }
        self._identity = identity
        self._evaluation_context.global_add('identity', self._identity)
        self._evaluation_context.global_context.merge(self._nested_items)

    @property
    def _nested_items(self) -> Dict[str, Aggregate]:
        """
        Dictionary of nested data groups
        """
        return self._aggregates

    def run_finalize(self) -> None:
        """
        Iteratively finalizes all data groups in its transformer
        """
        for item in self._nested_items.values():
            item.run_finalize()

    def __getattr__(self, item: str) -> Aggregate:
        """
        Makes the value of the nested items available as properties
        of the collection object.  This is used for retrieving data groups
        for dynamic execution.
        :param item: Aggregate requested
        """
        return self.__getitem__(item)

    def __getitem__(self, item) -> Aggregate:
        """
        Makes the nested items available though the square bracket notation.
        :raises KeyError: When a requested item is not found in nested items
        """
        if item not in self._nested_items:
            raise MissingAttributeError('{item} not defined in {name}'.format(
                item=item, name=self._name))

        return self._nested_items[item]



import json
from typing import List, Optional, Tuple, Dict, Union

from blurr.runner.data_processor import DataProcessor, SimpleJsonDataProcessor, \
    SimpleDictionaryDataProcessor
from blurr.runner.json_encoder import BlurrJSONEncoder
from blurr.runner.runner import Runner

_spark_import_err = None
try:
    from pyspark import RDD, SparkContext
    from pyspark.sql import SparkSession
except ImportError as err:
    # Ignore import error because the CLI can be used even if spark is not
    # installed.
    _spark_import_err = err

# Setting these at module level as they cannot be part of the spark runner object because they
# cannot be serialized
_module_spark_session: 'SparkSession' = None


def get_spark_session(spark_session: Optional['SparkSession'] = None) -> 'SparkSession':
    if spark_session:
        return spark_session

    global _module_spark_session
    if _module_spark_session:
        return _module_spark_session

    _module_spark_session = SparkSession \
        .builder \
        .appName("BlurrSparkRunner") \
        .getOrCreate()
    return _module_spark_session


class SparkRunner(Runner):
    """
    Provides these functionality in a Spark session:
    - Convert events data into Records.
    - Execute Blurr BTS on Records.
    - Basic output to file and to stdout.

    Example usage to process JSON events in files:
    ```
    json_files = [...]
    runner = SparkRunner(stream_bts_file, window_bts_file)
    # Provide a custom data processor if the json files contain data in a format other than
    # new-line separated JSONs where each JSON object is an event dictionary.
    records_rdd = runner.get_record_rdd_from_json_files(json_files)
    output_rdd = runner.execute(records_rdd)
    # Output rdd contains data in a Tuple[Identity, Tuple[Streaming BTS State, List of Window BTS
    # output]] format. These two can then be written out to external storage for passing BTS State
    # for the next execution run and as the training data for the model respectively.
    ```
    """

    def __init__(self, stream_bts_file: str, window_bts_file: Optional[str] = None):
        """
        Initialize SparkRunner.

        :param stream_bts_file: Streaming BTS to use. Must be provded.
        :param window_bts_file: Window BTS to use. If none is provided only the streaming BTS output
            is generated.
        """
        if _spark_import_err:
            raise _spark_import_err
        super().__init__(stream_bts_file, window_bts_file)

    def _execute_per_identity_records(
            self, identity_records_with_state: Tuple[str, Union[List, Tuple[List, Dict]]]):
        identity, records_with_state = identity_records_with_state
        if isinstance(records_with_state, tuple):
            record, state = records_with_state
        else:
            record, state = records_with_state, None
        return self.execute_per_identity_records(identity, record, state)

    def execute(self, identity_records: 'RDD', old_state_rdd: Optional['RDD'] = None) -> 'RDD':
        """
        Executes Blurr BTS with the given records. old_state_rdd can be provided to load an older
        state from a previous run.

        :param identity_records: RDD of the form Tuple[Identity, List[TimeAndRecord]]
        :param old_state_rdd: A previous streaming BTS state RDD as Tuple[Identity, Streaming BTS
            State]
        :return: RDD[Identity, Tuple[Streaming BTS State, List of Window BTS output]]
        """
        identity_records_with_state = identity_records
        if old_state_rdd:
            identity_records_with_state = identity_records.fullOuterJoin(old_state_rdd)
        return identity_records_with_state.map(lambda x: self._execute_per_identity_records(x))

    def get_record_rdd_from_json_files(self,
                                       json_files: List[str],
                                       data_processor: DataProcessor = SimpleJsonDataProcessor(),
                                       spark_session: Optional['SparkSession'] = None) -> 'RDD':
        """
        Reads the data from the given json_files path and converts them into the `Record`s format for
        processing. `data_processor` is used to process the per event data in those files to convert
        them into `Record`.

        :param json_files: List of json file paths. Regular Spark path wildcards are accepted.
        :param data_processor: `DataProcessor` to process each event in the json files.
        :param spark_session: `SparkSession` to use for execution. If None is provided then a basic
            `SparkSession` is created.
        :return: RDD containing Tuple[Identity, List[TimeAndRecord]] which can be used in
            `execute()`
        """
        spark_context = get_spark_session(spark_session).sparkContext
        raw_records: 'RDD' = spark_context.union(
            [spark_context.textFile(file) for file in json_files])
        return raw_records.mapPartitions(
            lambda x: self.get_per_identity_records(x, data_processor)).groupByKey().mapValues(list)

    def get_record_rdd_from_rdd(
            self,
            rdd: 'RDD',
            data_processor: DataProcessor = SimpleDictionaryDataProcessor(),
    ) -> 'RDD':
        """
        Converts a RDD of raw events into the `Record`s format for processing. `data_processor` is
        used to process the per row data to convert them into `Record`.

        :param rdd: RDD containing the raw events.
        :param data_processor: `DataProcessor` to process each row in the given `rdd`.
        :return: RDD containing Tuple[Identity, List[TimeAndRecord]] which can be used in
            `execute()`
        """
        return rdd.mapPartitions(
            lambda x: self.get_per_identity_records(x, data_processor)).groupByKey().mapValues(list)

    def write_output_file(self,
                          path: str,
                          per_identity_data: 'RDD',
                          spark_session: Optional['SparkSession'] = None) -> None:
        """
        Basic helper function to persist data to disk.

        If window BTS was provided then the window BTS output to written in csv format, otherwise,
        the streaming BTS output is written in JSON format to the `path` provided

        :param path: Path where the output should be written.
        :param per_identity_data: Output of the `execute()` call.
        :param spark_session: `SparkSession` to use for execution. If None is provided then a basic
            `SparkSession` is created.
        :return:
        """
        _spark_session_ = get_spark_session(spark_session)
        if not self._window_bts:
            per_identity_data.flatMap(
                lambda x: [json.dumps(data, cls=BlurrJSONEncoder) for data in x[1][0].items()]
            ).saveAsTextFile(path)
        else:
            # Convert to a DataFrame first so that the data can be saved as a CSV
            _spark_session_.createDataFrame(per_identity_data.flatMap(lambda x: x[1][1])).write.csv(
                path, header=True)

    def print_output(self, per_identity_data: 'RDD') -> None:
        """
        Basic helper function to write data to stdout. If window BTS was provided then the window
        BTS output is written, otherwise, the streaming BTS output is written to stdout.

        WARNING - For large datasets this will be extremely slow.

        :param per_identity_data: Output of the `execute()` call.
        """
        if not self._window_bts:
            data = per_identity_data.flatMap(
                lambda x: [json.dumps(data, cls=BlurrJSONEncoder) for data in x[1][0].items()])
        else:
            # Convert to a DataFrame first so that the data can be saved as a CSV
            data = per_identity_data.map(
                lambda x: json.dumps((x[0], x[1][1]), cls=BlurrJSONEncoder))
        for row in data.collect():
            print(row)



import json
from abc import ABC, abstractmethod
from typing import List, Dict

from blurr.core.record import Record


class DataProcessor(ABC):
    @abstractmethod
    def process_data(self, data_string: str) -> List[Record]:
        pass


class SimpleJsonDataProcessor(DataProcessor):
    def process_data(self, data_string: str) -> List[Record]:
        return [Record(json.loads(data_string))]


class SimpleDictionaryDataProcessor(DataProcessor):
    def process_data(self, data_dict: Dict) -> List[Record]:
        return [Record(data_dict)]


class IpfixDataProcessor(DataProcessor):
    IPFIX_EVENT_MAPPER = {
        4: 'protocol',
        12: 'dest_ip',
        56: 'source_mac',
        182: 'source_port',
        183: 'dest_port',
        150: 'timestamp'
    }

    def process_data(self, data_string: str) -> List[Record]:
        data = json.loads(data_string)
        if not isinstance(data, dict):
            return []

        record_list = []
        for data_row in data.get('DataSets', []):
            record = {}
            for event_dict in data_row:
                i = event_dict.get('I', 0)
                record[self.IPFIX_EVENT_MAPPER.get(i, i)] = event_dict['V']
            if self.IPFIX_EVENT_MAPPER[56] in record:
                record_list.append(Record(record))

        return record_list



from datetime import datetime, timezone
from typing import Any, Dict, List, Tuple

from dateutil import parser

from blurr.core.store import Store, Key, StoreSchema
from blurr.core.store_key import KeyType


class MemoryStoreSchema(StoreSchema):
    pass


class MemoryStore(Store):
    """
    In-memory store implementation
    """

    def __init__(self, schema: MemoryStoreSchema) -> None:
        self._schema = schema
        self._cache: Dict[Key, Any] = dict()

    def load(self):
        pass

    def get(self, key: Key) -> Any:
        return self._cache.get(key, None)

    def get_all(self, identity: str = None) -> Dict[Key, Any]:
        return {k: v
                for k, v in self._cache.items()
                if k.identity == identity} if identity else self._cache.copy()

    def _get_range_timestamp_key(self, start: Key, end: Key = None,
                                 count: int = 0) -> List[Tuple[Key, Any]]:
        items = []
        for key, item in self._cache.items():
            if start < key < end:
                items.append((key, item))
        items = sorted(items)
        if count:
            items = self._restrict_items_to_count(items, count)
        return items

    def _get_range_dimension_key(self,
                                 base_key: Key,
                                 start_time: datetime,
                                 end_time: datetime,
                                 count: int = 0) -> List[Tuple[Key, Any]]:
        start_time_str = start_time.isoformat()
        end_time_str = end_time.isoformat()
        items = []
        for key, item in self._cache.items():
            if key.starts_with(base_key):
                item_time = item.get('_start_time', datetime.min.isoformat())
                if start_time_str < item_time < end_time_str:
                    items.append((key, item))
        items = sorted(items, key=lambda i: i[1].get('_start_time', datetime.min.isoformat()))
        if count:
            items = self._restrict_items_to_count(items, count)
        return items

    def save(self, key: Key, item: Any) -> None:
        self._cache[key] = item

    def delete(self, key: Key) -> None:
        self._cache.pop(key, None)

    def finalize(self) -> None:
        pass



import importlib
from typing import Any, Union, Dict, Optional

from blurr.core.errors import TypeLoaderError
from blurr.core.type import Type

ITEM_MAP = {
    Type.BLURR_TRANSFORM_STREAMING: 'blurr.core.transformer_streaming.StreamingTransformer',
    Type.BLURR_TRANSFORM_WINDOW: 'blurr.core.transformer_window.WindowTransformer',
    Type.BLURR_AGGREGATE_BLOCK: 'blurr.core.aggregate_block.BlockAggregate',
    Type.BLURR_AGGREGATE_ACTIVITY: 'blurr.core.aggregate_activity.ActivityAggregate',
    Type.BLURR_AGGREGATE_IDENTITY: 'blurr.core.aggregate_identity.IdentityAggregate',
    Type.BLURR_AGGREGATE_VARIABLE: 'blurr.core.aggregate_variable.VariableAggregate',
    Type.BLURR_AGGREGATE_WINDOW: 'blurr.core.aggregate_window.WindowAggregate',
    Type.BLURR_STORE_MEMORY: 'blurr.store.memory_store.MemoryStore',
    Type.BLURR_STORE_DYNAMO: 'blurr.store.dynamo_store.DynamoStore',
    Type.DAY: 'blurr.core.window.Window',
    Type.HOUR: 'blurr.core.window.Window',
    Type.COUNT: 'blurr.core.window.Window',
    Type.STRING: 'blurr.core.field.Field',
    Type.INTEGER: 'blurr.core.field.Field',
    Type.BOOLEAN: 'blurr.core.field.Field',
    Type.DATETIME: 'blurr.core.field.Field',
    Type.FLOAT: 'blurr.core.field.Field',
    Type.MAP: 'blurr.core.field.Field',
    Type.LIST: 'blurr.core.field.Field',
    Type.SET: 'blurr.core.field.Field',
}

SCHEMA_MAP = {
    Type.BLURR_TRANSFORM_STREAMING: 'blurr.core.transformer_streaming.StreamingTransformerSchema',
    Type.BLURR_TRANSFORM_WINDOW: 'blurr.core.transformer_window.WindowTransformerSchema',
    Type.BLURR_AGGREGATE_BLOCK: 'blurr.core.aggregate_block.BlockAggregateSchema',
    Type.BLURR_AGGREGATE_ACTIVITY: 'blurr.core.aggregate_activity.ActivityAggregateSchema',
    Type.BLURR_AGGREGATE_IDENTITY: 'blurr.core.aggregate_identity.IdentityAggregateSchema',
    Type.BLURR_AGGREGATE_VARIABLE: 'blurr.core.aggregate_variable.VariableAggregateSchema',
    Type.BLURR_AGGREGATE_WINDOW: 'blurr.core.aggregate_window.WindowAggregateSchema',
    Type.BLURR_STORE_MEMORY: 'blurr.store.memory_store.MemoryStoreSchema',
    Type.BLURR_STORE_DYNAMO: 'blurr.store.dynamo_store.DynamoStoreSchema',
    Type.ANCHOR: 'blurr.core.anchor.AnchorSchema',
    Type.DAY: 'blurr.core.window.WindowSchema',
    Type.HOUR: 'blurr.core.window.WindowSchema',
    Type.COUNT: 'blurr.core.window.WindowSchema',
    Type.STRING: 'blurr.core.field_simple.StringFieldSchema',
    Type.INTEGER: 'blurr.core.field_simple.IntegerFieldSchema',
    Type.BOOLEAN: 'blurr.core.field_simple.BooleanFieldSchema',
    Type.DATETIME: 'blurr.core.field_simple.DateTimeFieldSchema',
    Type.FLOAT: 'blurr.core.field_simple.FloatFieldSchema',
    Type.MAP: 'blurr.core.field_complex.MapFieldSchema',
    Type.LIST: 'blurr.core.field_complex.ListFieldSchema',
    Type.SET: 'blurr.core.field_complex.SetFieldSchema'
}

# TODO Build dynamic type loader from a central configuration rather than reading a static dictionary

_class_cache: Dict[str, Type] = {}


class TypeLoader:
    @staticmethod
    def load_schema(type_name: Union[str, Type]):
        return TypeLoader.load_type(type_name, SCHEMA_MAP)

    @staticmethod
    def load_item(type_name: Union[str, Type]):
        return TypeLoader.load_type(type_name, ITEM_MAP)

    @staticmethod
    def load_type(type_name: Union[str, Type], type_map: dict) -> Any:
        type_class = None
        try:
            type_name_enum = Type(type_name)
            type_class = type_map[type_name_enum]
            return TypeLoader.import_class_by_full_name(type_map[type_name_enum])
        except (KeyError, ValueError):
            raise TypeLoaderError(str(type_name), type_class)

    @staticmethod
    def import_class_by_full_name(name) -> Optional[Type]:
        if name not in _class_cache:
            components = name.rsplit('.', 1)
            mod = importlib.import_module(components[0])
            _class_cache[name] = mod if len(components) == 1 else getattr(mod, components[1])
        return _class_cache.get(name, None)



from typing import Dict, Any, List, Optional, Union

from blurr.core.errors import BaseSchemaError, InvalidTypeError, TypeLoaderError, SpecNotFoundError
from blurr.core.errors import SchemaErrorCollection
from blurr.core.loader import TypeLoader
from blurr.core.type import Type
from blurr.core.validator import ATTRIBUTE_TYPE, ATTRIBUTE_NAME, validate_required_attributes


class SchemaLoader:
    """
    Provides functionality to operate on the schema using fully qualified names.
    """
    ITEM_SEPARATOR = '.'

    def __init__(self):
        self._spec_cache: Dict[str, Any] = {}
        self._schema_cache: Dict[str, 'BaseSchema'] = {}
        self._error_cache: SchemaErrorCollection = SchemaErrorCollection()
        self._store_cache: Dict[str, 'Store'] = {}

    def add_schema_spec(self, spec: Dict[str, Any],
                        fully_qualified_parent_name: str = None) -> Optional[str]:
        """
        Add a schema dictionary to the schema loader. The given schema is stored
        against fully_qualified_parent_name + ITEM_SEPARATOR('.') + schema.name.
        :param spec: Schema specification.
        :param fully_qualified_parent_name: Full qualified name of the parent.
        If None is passed then the schema is stored against the schema name.
        :return: The fully qualified name against which the spec is stored.
        None is returned if the given spec is not a dictionary or the spec does not
        contain a 'name' key.
        """
        if not isinstance(spec, dict) or ATTRIBUTE_NAME not in spec:
            return None

        name = spec[ATTRIBUTE_NAME]
        fully_qualified_name = name if fully_qualified_parent_name is None else self.get_fully_qualified_name(
            fully_qualified_parent_name, name)

        # Ensure that basic validation for each spec part is done before it is added to spec cache
        if isinstance(spec, dict):
            self._error_cache.add(
                validate_required_attributes(fully_qualified_name, spec, ATTRIBUTE_NAME,
                                             ATTRIBUTE_TYPE))
            if ATTRIBUTE_TYPE in spec and not Type.contains(spec[ATTRIBUTE_TYPE]):
                self._error_cache.add(
                    InvalidTypeError(fully_qualified_name, spec, ATTRIBUTE_TYPE,
                                     InvalidTypeError.Reason.TYPE_NOT_DEFINED))

        self._spec_cache[fully_qualified_name] = spec
        for key, val in spec.items():
            if isinstance(val, list):
                for item in val:
                    self.add_schema_spec(item, fully_qualified_name)
            self.add_schema_spec(val, fully_qualified_name)

        return spec[ATTRIBUTE_NAME]

    def add_errors(self, *errors: Union[BaseSchemaError, SchemaErrorCollection]) -> None:
        """ Adds errors to the error store for the schema """
        for error in errors:
            self._error_cache.add(error)

    def get_errors(self, fully_qualified_name: str = None,
                   include_nested: bool = True) -> List[BaseSchemaError]:
        if not fully_qualified_name:
            return self._error_cache.errors

        return self._error_cache[fully_qualified_name] + ([
            error for error in self._error_cache.errors
            if error.fully_qualified_name.startswith(fully_qualified_name + self.ITEM_SEPARATOR)
        ] if include_nested else [])

    def raise_errors(self) -> None:
        """ Raises errors that have been collected in the error cache """
        self._error_cache.raise_errors()

    # Using forward reference to avoid cyclic dependency.
    def get_schema_object(self, fully_qualified_name: str) -> 'BaseSchema':
        """
        Used to generate a schema object from the given fully_qualified_name.
        :param fully_qualified_name: The fully qualified name of the object needed.
        :return: An initialized schema object
        """

        if fully_qualified_name not in self._schema_cache:
            spec = self.get_schema_spec(fully_qualified_name)

            if spec:
                try:
                    self._schema_cache[fully_qualified_name] = TypeLoader.load_schema(
                        spec.get(ATTRIBUTE_TYPE, None))(fully_qualified_name, self)
                except TypeLoaderError as err:
                    self.add_errors(
                        InvalidTypeError(fully_qualified_name, spec, ATTRIBUTE_TYPE,
                                         InvalidTypeError.Reason.TYPE_NOT_LOADED,
                                         err.type_class_name))

        return self._schema_cache.get(fully_qualified_name, None)

    def get_store(self, fully_qualified_name: str) -> Optional['Store']:
        """
        Used to generate a store object from the given fully_qualified_name.
        :param fully_qualified_name: The fully qualified name of the store object needed.
        :return: An initialized store object
        """

        if fully_qualified_name not in self._store_cache:
            schema = self.get_schema_object(fully_qualified_name)
            if not schema:
                return None

            if Type.is_store_type(schema.type):
                self._store_cache[fully_qualified_name] = TypeLoader.load_item(schema.type)(schema)
            else:
                self.add_errors(
                    InvalidTypeError(fully_qualified_name, {}, ATTRIBUTE_TYPE,
                                     InvalidTypeError.Reason.INCORRECT_BASE, schema.type,
                                     InvalidTypeError.BaseTypes.STORE))

        return self._store_cache.get(fully_qualified_name, None)

    def get_all_stores(self) -> List['Store']:
        """
        Returns a list of stores that have been instantiated.
        """
        return list(self._store_cache.values())

    def get_nested_schema_object(self, fully_qualified_parent_name: str,
                                 nested_item_name: str) -> Optional['BaseSchema']:
        """
        Used to generate a schema object from the given fully_qualified_parent_name
        and the nested_item_name.
        :param fully_qualified_parent_name: The fully qualified name of the parent.
        :param nested_item_name: The nested item name.
        :return: An initialized schema object of the nested item.
        """
        return self.get_schema_object(
            self.get_fully_qualified_name(fully_qualified_parent_name, nested_item_name))

    @staticmethod
    def get_fully_qualified_name(fully_qualified_parent_name: str, nested_item_name: str) -> str:
        """
        Returns the fully qualified name by combining the fully_qualified_parent_name
        and nested_item_name.
        :param fully_qualified_parent_name: The fully qualified name of the parent.
        :param nested_item_name: The nested item name.
        :return: The fully qualified name of the nested item.
        """
        return fully_qualified_parent_name + SchemaLoader.ITEM_SEPARATOR + nested_item_name

    def get_schema_spec(self, fully_qualified_name: str) -> Dict[str, Any]:
        """
        Used to retrieve the specifications of the schema from the given
        fully_qualified_name of schema.
        :param fully_qualified_name: The fully qualified name of the schema needed.
        :return: Schema dictionary.
        """

        if fully_qualified_name not in self._spec_cache:
            self.add_errors(SpecNotFoundError(fully_qualified_name, {}))

        return self._spec_cache.get(fully_qualified_name, None)

    def has_schema_spec(self, fully_qualified_name: str) -> bool:
        """
        Checks if a schema spec exists in the schema loader
        :param fully_qualified_name: The fully qualified name of the schema needed.
        :return: True if spec definition exists, False otherwise.
        """

        return fully_qualified_name in self._spec_cache

    def get_schema_specs_of_type(self, *schema_types: Type) -> Dict[str, Dict[str, Any]]:
        """
        Returns a list of fully qualified names and schema dictionary tuples for
        the schema types provided.
        :param schema_types: Schema types.
        :return: List of fully qualified names and schema dictionary tuples.
        """

        return {
            fq_name: schema
            for fq_name, schema in self._spec_cache.items()
            if Type.is_type_in(schema.get(ATTRIBUTE_TYPE, ''), list(schema_types))
        }

    @staticmethod
    def get_transformer_name(fully_qualified_name: str) -> str:
        """
        Returns the BTS transformer name based on the given fully_qualified_name
        of one of the nested child items.
        :param fully_qualified_name: Fully qualified name of the nested child item.
        :return: BTS transformer name.
        """
        return fully_qualified_name.split(SchemaLoader.ITEM_SEPARATOR, 1)[0]



from abc import ABC, abstractmethod, abstractproperty
from typing import Dict, Type, Any

from blurr.core.base import BaseSchemaCollection, BaseItemCollection, BaseItem
from blurr.core.errors import MissingAttributeError
from blurr.core.evaluation import EvaluationContext
from blurr.core.field import Field
from blurr.core.loader import TypeLoader
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store import StoreSchema, Store
from blurr.core.store_key import Key, KeyType
from blurr.core.type import Type as BtsType
from blurr.core.validator import ATTRIBUTE_INTERNAL


class AggregateSchema(BaseSchemaCollection, ABC):
    """
    Group Schema must inherit from this base.  Data Group schema provides the
    abstraction for managing the 'Fields' in the group.
    """

    # Field Name Definitions
    ATTRIBUTE_STORE = 'Store'
    ATTRIBUTE_FIELDS = 'Fields'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        """
        Initializing the nested field schema that all data groups contain
        :param spec: Schema specifications for the field
        """
        super().__init__(fully_qualified_name, schema_loader, self.ATTRIBUTE_FIELDS)
        store_name = self._spec.get(self.ATTRIBUTE_STORE, None)
        self.store_schema: StoreSchema = None
        if store_name:
            self.store_schema = self.schema_loader.get_nested_schema_object(
                self.schema_loader.get_transformer_name(self.fully_qualified_name), store_name)

    def extend_schema_spec(self) -> None:
        """ Injects the identity field """
        super().extend_schema_spec()

        identity_field = {
            'Name': '_identity',
            'Type': BtsType.STRING,
            'Value': 'identity',
            ATTRIBUTE_INTERNAL: True
        }

        if self.ATTRIBUTE_FIELDS in self._spec:
            self._spec[self.ATTRIBUTE_FIELDS].insert(0, identity_field)
            self.schema_loader.add_schema_spec(identity_field, self.fully_qualified_name)


class Aggregate(BaseItemCollection, ABC):
    """
    All Data Groups inherit from this base.  Provides an abstraction for 'value' of the encapsulated
    to be called as properties of the data group itself.
    """

    def __init__(self, schema: AggregateSchema, identity: str,
                 evaluation_context: EvaluationContext) -> None:
        """
        Initializes the data group with the inherited context and adds
        its own nested items into the local context for execution
        :param schema: Schema for initializing the data group
        :param evaluation_context: Context dictionary for evaluation
        """
        super().__init__(schema, evaluation_context)
        self._identity = identity

        self._fields: Dict[str, Type[BaseItem]] = {
            name: TypeLoader.load_item(item_schema.type)(item_schema, self._evaluation_context)
            for name, item_schema in self._schema.nested_schema.items()
        }
        self._store: Store = None
        if self._schema.store_schema:
            self._store = self._schema.schema_loader.get_store(
                self._schema.store_schema.fully_qualified_name)

    @property
    def _nested_items(self) -> Dict[str, Field]:
        """
        Returns the dictionary of fields the Aggregate contains
        """
        return self._fields

    def run_finalize(self) -> None:
        """
        Saves the current state of the Aggregate in the store as the final rites
        """
        self._persist()

    @property
    def _key(self):
        return Key(KeyType.DIMENSION, self._identity, self._name)

    def _persist(self) -> None:
        """
        Persists the current data group
        """
        if self._store:
            self._store.save(self._key, self._snapshot)

    def __getattr__(self, item: str) -> Any:
        """
        Makes the value of the nested items available as properties
        of the collection object.  This is used for retrieving field values
        for dynamic execution.
        :param item: Field requested
        """
        return self.__getitem__(item)

    def __getitem__(self, item: str) -> Any:
        """
        Makes the nested items available though the square bracket notation.
        :raises KeyError: When a requested item is not found in nested items
        """
        if item not in self._nested_items:
            raise MissingAttributeError('{item} not defined in {name}'.format(
                item=item, name=self._name))

        return self._nested_items[item].value



from collections import defaultdict
from datetime import datetime
from typing import Dict

from blurr.core.aggregate_block import BlockAggregate, TimeAggregate
from blurr.core.base import BaseSchema
from blurr.core.evaluation import EvaluationContext
from blurr.core.schema_loader import SchemaLoader


class AnchorSchema(BaseSchema):
    """
    Represents the schema for the Anchor specified in a window BTS.
    """

    ATTRIBUTE_CONDITION = 'Condition'
    ATTRIBUTE_MAX = 'Max'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.condition = self.build_expression(self.ATTRIBUTE_CONDITION)
        self.max = self._spec[self.ATTRIBUTE_MAX] if self.ATTRIBUTE_MAX in self._spec else None

    def validate_schema_spec(self) -> None:
        self.validate_required_attributes(self.ATTRIBUTE_CONDITION)
        self.validate_number_attribute(self.ATTRIBUTE_MAX, int, 1)


class Anchor:
    def __init__(self, schema: AnchorSchema):
        self._schema = schema
        self._condition_met: Dict[datetime, int] = defaultdict(int)
        self.anchor_block = None

    def evaluate_anchor(self, block: TimeAggregate, evaluation_context: EvaluationContext) -> bool:
        self.anchor_block = block
        if self.max_condition_met(block):
            return False

        if self._schema.condition.evaluate(evaluation_context):
            return True

        return False

    def add_condition_met(self):
        self._condition_met[self.anchor_block._start_time.date()] += 1

    def max_condition_met(self, block: TimeAggregate) -> bool:
        if self._schema.max is None:
            return False
        return self._condition_met[block._start_time.date()] >= self._schema.max



import sys
from typing import List, Any, Dict

import yaml

from blurr.cli.util import get_yml_files, eprint
from blurr.core import logging
from blurr.core.errors import SchemaError, InvalidSpecError
from blurr.core.schema_loader import SchemaLoader


def validate_command(bts_files: List[str]) -> int:
    all_files_valid = True
    if len(bts_files) == 0:
        bts_files = get_yml_files()
    for bts_file in bts_files:
        if validate_file(bts_file) == 1:
            all_files_valid = False

    return 0 if all_files_valid else 1


def validate_file(bts_file: str) -> int:
    print('Running validation on {}'.format(bts_file))
    try:
        bts_dict = yaml.safe_load(open(bts_file, 'r', encoding='utf-8'))
        validate(bts_dict)
        print('Document is valid')
        return 0
    except yaml.YAMLError as err:
        eprint('Invalid yaml')
        eprint(str(err))
        return 1
    except SchemaError as err:
        eprint(str(err))
        return 1
    except Exception as err:
        exception_value = sys.exc_info()[1]
        logging.error(exception_value)
        eprint('There was an error parsing the document. Error:\n' + str(err))
        return 1


def validate(spec: Dict[str, Any]) -> None:
    schema_loader = SchemaLoader()
    bts_name = schema_loader.add_schema_spec(spec)
    if not bts_name:
        raise InvalidSpecError(spec)
    schema_loader.raise_errors()
    schema_loader.get_schema_object(bts_name)
    print(schema_loader.get_errors())
    schema_loader.raise_errors()


def get_valid_yml_files(yml_files: List[str]) -> List[str]:
    return [yml_file for yml_file in yml_files if validate_file(yml_file) == 0]



from datetime import datetime, timezone
from typing import Any

from dateutil import parser

from blurr.core.field import FieldSchema


class IntegerFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return int

    @property
    def default(self) -> Any:
        return int(0)


class FloatFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return float

    @property
    def default(self) -> Any:
        return float(0)


class StringFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return str

    @property
    def default(self) -> Any:
        return str()


class BooleanFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return bool

    @property
    def default(self) -> Any:
        return False


class DateTimeFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return datetime

    @property
    def default(self) -> Any:
        return None

    @staticmethod
    def sanitize_object(instance: datetime) -> datetime:
        if instance.tzinfo is None or instance.tzinfo.utcoffset(instance) is None:
            instance = instance.replace(tzinfo=timezone.utc)

        return instance

    @staticmethod
    def encoder(value: Any) -> str:
        return value.isoformat() if value else None

    def decoder(self, value: Any) -> datetime:
        return self.sanitize_object(parser.parse(value)) if value else None



from datetime import datetime, timedelta
from typing import Any, List, Tuple

from blurr.core.aggregate import Aggregate, AggregateSchema
from blurr.core.aggregate_block import BlockAggregate, BlockAggregateSchema, TimeAggregate
from blurr.core.aggregate_time import TimeAggregateSchema
from blurr.core.errors import PrepareWindowMissingBlocksError
from blurr.core.evaluation import EvaluationContext
from blurr.core.loader import TypeLoader
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store_key import Key, KeyType
from blurr.core.type import Type


class WindowAggregateSchema(AggregateSchema):

    ATTRIBUTE_WINDOW_VALUE = 'WindowValue'
    ATTRIBUTE_WINDOW_TYPE = 'WindowType'
    ATTRIBUTE_SOURCE = 'Source'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.window_value = self._spec.get(self.ATTRIBUTE_WINDOW_VALUE, 0)
        self.window_type = self._spec.get(self.ATTRIBUTE_WINDOW_TYPE, None)

        self.source: TimeAggregateSchema = self.schema_loader.get_schema_object(self._spec[self.ATTRIBUTE_SOURCE]) if \
            self.ATTRIBUTE_SOURCE in self._spec and schema_loader.has_schema_spec(self._spec[self.ATTRIBUTE_SOURCE]) \
            else None

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_WINDOW_TYPE, self.ATTRIBUTE_WINDOW_VALUE,
                                          self.ATTRIBUTE_SOURCE)
        self.validate_number_attribute(self.ATTRIBUTE_WINDOW_VALUE, int)


class _WindowSource:
    """
    Represents a window on the pre-aggregated source data.
    """

    def __init__(self, block_list: List[TimeAggregate]):
        self.view: List[TimeAggregate] = block_list

    def __getattr__(self, item: str) -> List[Any]:
        return [getattr(block, item) for block in self.view]


class WindowAggregate(Aggregate):
    """
    Manages the generation of WindowAggregate as defined in the schema.
    """

    def __init__(self, schema: WindowAggregateSchema, identity: str,
                 evaluation_context: EvaluationContext) -> None:
        super().__init__(schema, identity, evaluation_context)
        self._window_source = None

    def _prepare_window(self, start_time: datetime) -> None:
        """
        Prepares window if any is specified.
        :param start_time: The anchor block start_time from where the window
        should be generated.
        """
        # evaluate window first which sets the correct window in the store
        store = self._schema.schema_loader.get_store(
            self._schema.source.store_schema.fully_qualified_name)
        if Type.is_type_equal(self._schema.window_type, Type.DAY) or Type.is_type_equal(
                self._schema.window_type, Type.HOUR):
            block_list = self._load_blocks(
                store.get_range(
                    Key(self._schema.source.key_type, self._identity, self._schema.source.name),
                    start_time, self._get_end_time(start_time)))
        else:
            block_list = self._load_blocks(
                store.get_range(
                    Key(self._schema.source.key_type, self._identity, self._schema.source.name),
                    start_time, None, self._schema.window_value))

        self._window_source = _WindowSource(block_list)
        self._validate_view()

    def _validate_view(self):
        if Type.is_type_equal(
                self._schema.window_type,
                Type.COUNT) and len(self._window_source.view) != abs(self._schema.window_value):
            raise PrepareWindowMissingBlocksError(
                '{} WindowAggregate: Expecting {} but found {} blocks'.format(
                    self._schema.name, abs(self._schema.window_value),
                    len(self._window_source.view)))

        if len(self._window_source.view) == 0:
            raise PrepareWindowMissingBlocksError(
                '{} WindowAggregate: No matching blocks found'.format(self._schema.name))

    # TODO: Handle end time which is beyond the expected range of data being
    # processed. In this case a PrepareWindowMissingBlocksError error should
    # be raised.
    def _get_end_time(self, start_time: datetime) -> datetime:
        """
        Generates the end time to be used for the store range query.
        :param start_time: Start time to use as an offset to calculate the end time
        based on the window type in the schema.
        :return:
        """
        if Type.is_type_equal(self._schema.window_type, Type.DAY):
            return start_time + timedelta(days=self._schema.window_value)
        elif Type.is_type_equal(self._schema.window_type, Type.HOUR):
            return start_time + timedelta(hours=self._schema.window_value)

    def _load_blocks(self, blocks: List[Tuple[Key, Any]]) -> List[TimeAggregate]:
        """
        Converts [(Key, block)] to [BlockAggregate]
        :param blocks: List of (Key, block) blocks.
        :return: List of BlockAggregate
        """
        return [
            TypeLoader.load_item(self._schema.source.type)(self._schema.source, self._identity,
                                                           EvaluationContext()).run_restore(block)
            for (_, block) in blocks
        ]

    def run_evaluate(self) -> None:
        self._evaluation_context.local_context.add('source', self._window_source)
        super().run_evaluate()



from abc import ABC, abstractmethod
from typing import Dict, Any, Type, TypeVar, Union, List, Optional, Set

from blurr.core.errors import SnapshotError, BaseSchemaError, InvalidExpressionError
from blurr.core.evaluation import Expression, EvaluationContext
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store_key import Key
from blurr.core.validator import validate_required_attributes, \
    validate_python_identifier_attributes, \
    validate_number_attribute, validate_empty_attributes, validate_enum_attribute


class BaseSchema(ABC):
    """ Encapsulates the common functionality of all schema elements """

    ATTRIBUTE_NAME = 'Name'
    ATTRIBUTE_DESCRIPTION = 'Description'
    ATTRIBUTE_TYPE = 'Type'
    ATTRIBUTE_WHEN = 'When'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        """
        Initializes a schema by providing the path to the schema and the schema loader resource
        :param fully_qualified_name: Fully qualified path to the schema
        :param schema_loader: Schema repository that returns schema spec by fully qualified name
        """
        self.schema_loader: SchemaLoader = schema_loader
        self.fully_qualified_name: str = fully_qualified_name
        self._spec: Dict[str, Any] = self.schema_loader.get_schema_spec(self.fully_qualified_name)

        self.validate_schema_spec()

        self.extend_schema_spec()

        self.name: str = self._spec[self.ATTRIBUTE_NAME]
        self.type: str = self._spec[self.ATTRIBUTE_TYPE]

        self.when: Expression = self.build_expression(self.ATTRIBUTE_WHEN)
        self.description: str = self._spec.get(self.ATTRIBUTE_DESCRIPTION, None)

    def build_expression(self, attribute: str) -> Optional[Expression]:
        """ Builds an expression object.  Adds an error if expression creation has errors. """

        expression_string = self._spec.get(attribute, None)
        if expression_string:
            try:
                return Expression(str(expression_string))
            except Exception as err:
                self.add_errors(
                    InvalidExpressionError(self.fully_qualified_name, self._spec, attribute, err))

        return None

    def add_errors(self, *errors: Union[BaseSchemaError, List[BaseSchemaError]]) -> None:
        """ Adds errors to the error repository in schema loader """
        self.schema_loader.add_errors(*errors)

    @property
    def errors(self) -> List[BaseSchemaError]:
        """ Returns a list of errors raised by this schema """
        return self.schema_loader.get_errors(self.fully_qualified_name)

    def validate_required_attributes(self, *attributes: str) -> None:
        """ Validates that the schema contains a series of required attributes """
        self.add_errors(
            validate_required_attributes(self.fully_qualified_name, self._spec, *attributes))

    def validate_number_attribute(self,
                                  attribute: str,
                                  value_type: Union[Type[int], Type[float]] = int,
                                  minimum: Optional[Union[int, float]] = None,
                                  maximum: Optional[Union[int, float]] = None) -> None:
        """ Validates that the attribute contains a numeric value within boundaries if specified """
        self.add_errors(
            validate_number_attribute(self.fully_qualified_name, self._spec, attribute, value_type,
                                      minimum, maximum))

    def validate_enum_attribute(self, attribute: str,
                                candidates: Set[Union[str, int, float]]) -> None:
        """ Validates that the attribute value is among the candidates """
        self.add_errors(
            validate_enum_attribute(self.fully_qualified_name, self._spec, attribute, candidates))

    def validate_schema_spec(self) -> None:
        """ Contains the validation routines that are to be executed as part of initialization by subclasses.
        When this method is being extended, the first line should always be: ```super().validate_schema_spec()``` """
        self.add_errors(
            validate_empty_attributes(self.fully_qualified_name, self._spec, *self._spec.keys()))
        self.add_errors(
            validate_python_identifier_attributes(self.fully_qualified_name, self._spec,
                                                  self.ATTRIBUTE_NAME))

    def extend_schema_spec(self) -> None:
        """ Extends the defined schema specifications at runtime with defaults. When this method is being extended,
        the first line should always be:  ```super().extend_schema_spec()``` """
        pass


class BaseSchemaCollection(BaseSchema, ABC):
    """
    Base class for schema that contain nested schema
    """

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader,
                 nested_schema_attribute: str) -> None:
        """
        Initializes a schema collection with nested items
        :param fully_qualified_name: Fully qualified path to the schema
        :param schema_loader: Schema repository that returns schema spec by fully qualified name
        :param nested_schema_attribute: Name of the attribute that contains the nested elements
        """
        # Nested Items must be initialized before base init so that the validation routine finds it
        self._nested_item_attribute = nested_schema_attribute

        super().__init__(fully_qualified_name, schema_loader)

        # Load nested schema items
        self.nested_schema: Dict[str, Type[BaseSchema]] = {
            schema_spec[self.ATTRIBUTE_NAME]: self.schema_loader.get_nested_schema_object(
                self.fully_qualified_name, schema_spec[self.ATTRIBUTE_NAME])
            for schema_spec in self._spec.get(self._nested_item_attribute, [])
        }

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self._nested_item_attribute)


BaseItemType = TypeVar('T', bound='BaseItem')


class BaseItem(ABC):
    """
    Base class for for all leaf items that do not contain sub-items
    """

    def __init__(self, schema: BaseSchema, evaluation_context: EvaluationContext) -> None:
        """
        Initializes an item with the schema and execution context
        :param schema: Schema of the item
        :param evaluation_context: Context dictionary for evaluation
        """
        self._schema = schema
        self._evaluation_context: EvaluationContext = evaluation_context

    @property
    def _needs_evaluation(self) -> bool:
        """
        Returns True when:
            1. Where clause is not specified
            2. Where WHERE clause is specified and it evaluates to True
        Returns false if a where clause is specified and it evaluates to False
        """
        return self._schema.when is None or self._schema.when.evaluate(self._evaluation_context)

    @property
    def _name(self) -> str:
        """
        Returns the name of the base item
        """
        return self._schema.name

    @abstractmethod
    def run_evaluate(self, *args, **kwargs) -> None:
        """
        Evaluates the current item
        """
        raise NotImplementedError('evaluate() must be implemented')

    @property
    @abstractmethod
    def _snapshot(self):
        """
        Gets a dictionary representation of the current state items in the current hierarchy
        :return: Name, Value map of the current tree
        """
        raise NotImplementedError('snapshot() must be implemented')

    @abstractmethod
    def run_restore(self, snapshot) -> BaseItemType:
        """
        Restores the state of an item from a snapshot
        """
        raise NotImplementedError('restore() must be implemented')

    @abstractmethod
    def run_reset(self) -> None:
        """
        Resets the state of the item.
        """
        raise NotImplementedError('reset() must be implemented')


class BaseItemCollection(BaseItem, ABC):
    """
    Base class for items that contain sub-items within them
    """

    def __init__(self, schema: BaseSchemaCollection, evaluation_context: EvaluationContext) -> None:
        """
        Loads nested items to the 'items' collection
        :param schema: Schema that conforms to the item
        :param evaluation_context: Context dictionary for evaluation
        """

        super().__init__(schema, evaluation_context)

    def run_evaluate(self, *args, **kwargs) -> None:
        """
        Evaluates the current item
        :returns An evaluation result object containing the result, or reasons why
        evaluation failed
        """
        if self._needs_evaluation:
            for _, item in self._nested_items.items():
                item.run_evaluate()

    @property
    def _snapshot(self) -> Dict[str, Any]:
        """
        Implements snapshot for collections by recursively invoking snapshot of all child items
        """
        try:
            return {name: item._snapshot for name, item in self._nested_items.items()}
        except Exception as e:
            raise SnapshotError('Error while creating snapshot for {}'.format(self._name)) from e

    def run_restore(self, snapshot: Dict[Union[str, Key], Any]) -> 'BaseItemCollection':
        """
        Restores the state of a collection from a snapshot
        """
        try:

            for name, snap in snapshot.items():
                if isinstance(name, Key):
                    self._nested_items[name.group].run_restore(snap)
                else:
                    self._nested_items[name].run_restore(snap)
            return self

        except Exception as e:
            raise SnapshotError('Error while restoring snapshot: {}'.format(self._snapshot)) from e

    def run_reset(self) -> None:
        for _, item in self._nested_items.items():
            item.run_reset()

    @abstractmethod
    def run_finalize(self) -> None:
        """
        Performs the final rites of an item before it decommissioned
        """
        raise NotImplementedError('finalize() must be implemented')

    @property
    @abstractmethod
    def _nested_items(self) -> Dict[str, Type[BaseItem]]:
        """
        Dictionary of the name and item in the collection
        """
        raise NotImplementedError('nested_items() must be implemented')



from typing import Any

from blurr.core.field import FieldSchema, ComplexTypeBase


class Map(ComplexTypeBase, dict):
    """
    Extends native dictionary with operations for evaluation support.
    """

    def set(self, key: Any, value: Any) -> None:
        """ Sets the value of a key to a supplied value """
        if key is not None:
            self[key] = value

    def increment(self, key: Any, by: int = 1) -> None:
        """ Increments the value set against a key.  If the key is not present, 0 is assumed as the initial state """
        if key is not None:
            self[key] = self.get(key, 0) + by


class MapFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return Map

    @property
    def default(self) -> Any:
        return Map()


class List(ComplexTypeBase, list):
    """
    Extends native list with operations for evaluation support.
    """

    def append(self, obj: Any) -> None:
        """ Appends an object to the list as long as it is not None """
        if obj is not None:
            super().append(obj)

    def insert(self, index: int, obj: Any) -> None:
        """ Inserts an item to the list as long as it is not None """
        if obj is not None:
            super().insert(index, obj)


class ListFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return List

    @property
    def default(self) -> Any:
        return List()


class Set(ComplexTypeBase, set):
    """
    Extends native set with operations for evaluation support.
    """

    def add(self, element: Any) -> None:
        """ Adds an element to the set as long as it is not None """
        if element is not None:
            super().add(element)


class SetFieldSchema(FieldSchema):
    @property
    def type_object(self) -> Any:
        return Set

    @property
    def default(self) -> Any:
        return Set()

    @staticmethod
    def encoder(value: Any) -> Set:
        return list(value)



from typing import Dict, Optional

from blurr.core.aggregate_block import BlockAggregate, TimeAggregate
from blurr.core.aggregate_window import WindowAggregate
from blurr.core.anchor import Anchor
from blurr.core.errors import AnchorBlockNotDefinedError
from blurr.core.evaluation import Context, EvaluationContext
from blurr.core.schema_loader import SchemaLoader
from blurr.core.transformer import Transformer, TransformerSchema
from blurr.core.type import Type


class WindowTransformerSchema(TransformerSchema):
    """
    Represents the schema for processing aggregated data using windows.
    Handles attributes specific to the window BTS schema
    """

    ATTRIBUTE_ANCHOR = 'Anchor'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.anchor = self.schema_loader.get_schema_object(
            self.fully_qualified_name + '.anchor') if self.ATTRIBUTE_ANCHOR in self._spec else None

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_ANCHOR)

    def extend_schema_spec(self) -> None:
        super().extend_schema_spec()
        if not self.errors:
            self._spec[self.ATTRIBUTE_ANCHOR][self.ATTRIBUTE_NAME] = 'anchor'
            self._spec[self.ATTRIBUTE_ANCHOR][self.ATTRIBUTE_TYPE] = Type.ANCHOR
            self.schema_loader.add_schema_spec(self._spec[self.ATTRIBUTE_ANCHOR],
                                               self.fully_qualified_name)


class WindowTransformer(Transformer):
    """
    The Window BTS transformer that performs window operations on pre-aggregated
    block data.
    """

    def __init__(self, schema: WindowTransformerSchema, identity: str, context: Context) -> None:
        super().__init__(schema, identity)
        self._evaluation_context.merge(EvaluationContext(context))
        self._anchor = Anchor(schema.anchor)

    def run_evaluate(self, block: TimeAggregate) -> bool:
        """
        Evaluates the anchor condition against the specified block.
        :param block: Block to run the anchor condition against.
        :return: True, if the anchor condition is met, otherwise, False.
        """
        if self._anchor.evaluate_anchor(block, self._evaluation_context):

            try:
                self.run_reset()
                self._evaluation_context.global_add('anchor', block)
                self._evaluate()
                self._anchor.add_condition_met()
                return True
            finally:
                self._evaluation_context.global_remove('anchor')

        return False

    def _evaluate(self):
        if 'anchor' not in self._evaluation_context.global_context or self._anchor.anchor_block is None:
            raise AnchorBlockNotDefinedError()

        if not self._needs_evaluation:
            return

        for item in self._nested_items.values():
            if isinstance(item, WindowAggregate):
                item._prepare_window(self._anchor.anchor_block._start_time)

        super().run_evaluate()

    @property
    def run_flattened_snapshot(self) -> Dict:
        """
        Generates a flattened snapshot where the final key for a field is <aggregate_name>.<field_name>.
        :return: The flattened snapshot.
        """
        snapshot_dict = super()._snapshot

        # Flatten to feature dict
        return self._flatten_snapshot(None, snapshot_dict)

    def _flatten_snapshot(self, prefix: Optional[str], value: Dict) -> Dict:
        flattened_dict = {}
        for k, v in value.items():
            if isinstance(v, dict):
                flattened_dict.update(self._flatten_snapshot(k, v))
            else:
                flattened_dict[prefix + '.' + k if prefix is not None else k] = v

        return flattened_dict



from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum
from io import StringIO
from itertools import chain
from os import linesep
from typing import List, Dict, Any, Union, Type, Set, Tuple


class GenericSchemaError(Exception):
    pass


class BaseSchemaError(Exception, ABC):
    """
    Indicates an error in the schema specification
    """

    def __init__(self, fully_qualified_name: str, spec: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fully_qualified_name = fully_qualified_name
        self.spec = spec

    def __repr__(self):
        return '{cls}: FQN: {fqn}'.format(
            cls=self.__class__.__name__, fqn=self.fully_qualified_name)

    @property
    @abstractmethod
    def key(self) -> Tuple:
        """ Returns a tuple that uniquely identifies the object by its values """
        return (self.fully_qualified_name, )

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return type(self) == type(other) and self.key == other.key


class BaseSchemaAttributeError(BaseSchemaError, ABC):
    """
    Indicates an error in the schema specification
    """

    def __init__(self, fully_qualified_name: str, spec: Dict[str, Any], attribute: str, *args,
                 **kwargs):
        super().__init__(fully_qualified_name, spec, *args, **kwargs)
        self.attribute = attribute

    def __repr__(self):
        return '{cls}: FQN: {fqn}, Attribute: {attribute}'.format(
            cls=self.__class__.__name__, fqn=self.fully_qualified_name, attribute=self.attribute)

    @property
    def key(self):
        return super().key + (self.attribute, )


class RequiredAttributeError(BaseSchemaAttributeError):
    def __str__(self):
        return 'Attribute `{}` must be present under `{}`.'.format(self.attribute,
                                                                   self.fully_qualified_name)


class EmptyAttributeError(BaseSchemaAttributeError):
    def __str__(self):
        return 'Attribute `{}` under `{}` cannot be left empty.'.format(
            self.attribute, self.fully_qualified_name)


class InvalidValueError(BaseSchemaAttributeError):
    def __init__(self, fully_qualified_name: str, spec: Dict[str, Any], attribute: str,
                 candidates: Set[Any], *args, **kwargs):
        super().__init__(fully_qualified_name, spec, attribute, *args, **kwargs)
        self.candidates = candidates

    def __str__(self):
        return 'Attribute `{attr}` under `{fqn}` must have one of the following values: {candidates}'.format(
            attr=self.attribute,
            fqn=self.fully_qualified_name,
            candidates=' | '.join([str(x) for x in self.candidates]))

    @property
    def key(self):
        return super().key + (str(self.candidates), )


class InvalidNumberError(BaseSchemaAttributeError):
    def __init__(self,
                 fully_qualified_name: str,
                 spec: Dict[str, Any],
                 attribute: str,
                 value_type: Type,
                 minimum: Any = None,
                 maximum: Any = None,
                 *args,
                 **kwargs):
        super().__init__(fully_qualified_name, spec, attribute, *args, **kwargs)
        self.type = value_type
        self.min = minimum
        self.max = maximum

    def __str__(self):
        return 'Attribute `{attr}` under `{fqn}` must be of type `{type}`. {less_than} {greater_than}'.format(
            attr=self.attribute,
            fqn=self.fully_qualified_name,
            type=self.type.__name__,
            greater_than=('Must be greater than ' + str(self.min)) if self.min else '',
            less_than=('Must be lesser than ' + str(self.max)) if self.max else '')

    @property
    def key(self):
        return super().key + (self.type.__name__, self.min, self.max)


class InvalidIdentifierError(BaseSchemaAttributeError):
    class Reason(Enum):
        STARTS_WITH_UNDERSCORE = 'Identifiers starting with underscore `_` are reserved'
        STARTS_WITH_RUN = 'Identifiers starting with `run_` are reserved'
        INVALID_PYTHON_IDENTIFIER = 'Identifiers must be valid Python identifiers'

    def __init__(self, fully_qualified_name: str, spec: Dict[str, Any], attribute: str,
                 reason: 'InvalidIdentifierError.Reason', *args, **kwargs):
        super().__init__(fully_qualified_name, spec, attribute, *args, **kwargs)
        self.reason = reason

    def __str__(self):
        return '`{attribute}: {value}` in section `{name}` is invalid. {reason}.'.format(
            attribute=self.attribute,
            value=self.spec.get(self.attribute, '*missing*'),
            name=self.fully_qualified_name,
            reason=self.reason.value)

    @property
    def key(self):
        return super().key + (str(self.reason), )


class InvalidTypeError(BaseSchemaAttributeError):
    class Reason(Enum):
        TYPE_NOT_DEFINED = 'Type `{type_name}` is not declared in the system configuration.'
        TYPE_NOT_LOADED = 'Class `{type_class_name}` could not be loaded.'
        INCORRECT_BASE = 'Object does not inherit from the expected base class {expected_base_type}.'

    class BaseTypes:
        SCHEMA = 'BaseSchema'
        ITEM = 'BaseItem'
        STORE = 'Store'

    def __init__(self,
                 fully_qualified_name: str,
                 spec: Dict[str, Any],
                 attribute: str,
                 reason: 'InvalidTypeError.Reason',
                 type_class_name: str = None,
                 expected_base_type: BaseTypes = None,
                 *args,
                 **kwargs):
        super().__init__(fully_qualified_name, spec, attribute, *args, **kwargs)
        self.reason = reason
        self.type_class_name = type_class_name
        self.expected_base_type = expected_base_type

    def __str__(self):
        return '`{attribute}: {value}` in section `{name}` is invalid. {reason}.'.format(
            attribute=self.attribute,
            value=self.spec.get(self.attribute, '*missing*'),
            name=self.fully_qualified_name,
            reason=self.reason.value.format(
                type_name=self.spec.get(self.attribute, '*missing*'),
                expected_base_type=self.expected_base_type.value,
                type_class_name=self.type_class_name))

    @property
    def key(self):
        return super().key + (str(self.reason), str(self.expected_base_type), self.type_class_name)


class InvalidExpressionError(BaseSchemaAttributeError):
    """
    Indicates that a python expression specified is either non-compilable, or not allowed
    """

    def __init__(self, fully_qualified_name: str, spec: Dict[str, Any], attribute: str,
                 error: Exception, *args, **kwargs):
        super().__init__(fully_qualified_name, spec, attribute, *args, **kwargs)
        self.error = error

    def __str__(self):
        return '`{attribute}: {value}` in section `{name}` is invalid Python expression. Compilation error: \n{error}'.format(
            attribute=self.attribute,
            value=self.spec.get(self.attribute, '*missing*'),
            name=self.fully_qualified_name,
            error=str(self.error))


class SchemaErrorCollection:
    def __init__(self, *args):
        self.log: Dict[str, Set[BaseSchemaError]] = defaultdict(set)
        for arg in args:
            self.add(arg)

    def add(self, item: Union[BaseSchemaError, List[BaseSchemaError]]):
        if isinstance(item, BaseSchemaError):
            self.log[item.fully_qualified_name].add(item)

        elif isinstance(item, list):
            for i in item:
                self.add(i)

    def merge(self, item: 'SchemaErrorCollection'):
        if not item:
            return

        for k, v in item.log.items():
            self.log[k].update(v)

    def __str__(self):
        return linesep.join(
            [str(error) for error in self.log.values()]) if len(self.log) > 0 else ''

    def __getitem__(self, item):
        return list(self.log[item])

    def __contains__(self, item):
        return self.log.__contains__(item)

    def __iter__(self):
        return iter(self.log.items())

    @property
    def errors(self) -> List[BaseSchemaError]:
        return list(chain.from_iterable(self.log.values()))

    @property
    def has_errors(self) -> bool:
        return len(self.errors) > 0

    def raise_errors(self):
        if self.has_errors:
            raise SchemaError(self)


class SchemaErrorCollectionFormatter:
    def __init__(self, **kwargs):
        self.header_separator = kwargs.get('header_separator', '=')
        self.error_separator = kwargs.get('item_separator', '-')
        self.line_separator = kwargs.get('line_separator', linesep)

    def format(self, errors: SchemaErrorCollection) -> Any:
        with StringIO() as result:
            for fqn, errs in errors:
                result.writelines([
                    self.line_separator, fqn, self.line_separator, self.header_separator * len(fqn),
                    self.line_separator
                ])
                for err in errs:
                    result.writelines(['--> ', str(err), self.line_separator])

            return result.getvalue()


class SchemaError(Exception):
    def __init__(self, errors: SchemaErrorCollection, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.errors = errors
        self.formatter = SchemaErrorCollectionFormatter()

    def __str__(self):
        return self.formatter.format(self.errors)

    def __repr__(self):
        return self.__class__.__name__ + linesep + str(self)


class SpecNotFoundError(BaseSchemaError):
    @property
    def key(self):
        return super().key


class InvalidSpecError(BaseSchemaError):
    def __init__(self, spec: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fully_qualified_name = '**InvalidSpec**'
        self.spec = spec

    @property
    def key(self):
        return super().key

    def __str__(self):
        return 'The following spec is invalid: \n{spec}'.format(spec=self.spec)


class ExpressionEvaluationError(Exception):
    """
    Error raised during expression evaluation by the interpreter
    """
    pass


class TypeLoaderError(Exception):
    """
    Indicates dynamic type loading failure
    """

    def __init__(self, type_name: str = '', type_class_name: str = '', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type_name = type_name
        self.type_class_name = type_class_name

    def __str__(self):
        return 'Failed to load class `{type_class_name}` of type `{type_name}`.'.format(
            type_class_name=self.type_class_name, type_name=self.type_name)


class SnapshotError(Exception):
    """
    Indicates issues with serializing the current state of the object
    """
    pass


class StaleBlockError(Exception):
    """
    Indicates that the event being processed cannot be added to the block rollup that is loaded
    """
    pass


class StreamingSourceNotFoundError(Exception):
    """
    Raised when the raw data for streaming is unavailable in the execution context
    """
    pass


class AnchorBlockNotDefinedError(Exception):
    """
    Raised when anchor block is not defined and a WindowTransformer is evaluated.
    """
    pass


class IdentityError(Exception):
    """
    Raised when there is an error in the identity determination of a record.
    """
    pass


class TimeError(Exception):
    """
    Raised when there is an error in determining the time of the record.
    """
    pass


class PrepareWindowMissingBlocksError(Exception):
    """
    Raised when the window view generated is insufficient as per the window specification.
    """
    pass


class MissingAttributeError(Exception):
    """
    Raised when the name of the item being retrieved does not exist in the nested items.
    """
    pass


class KeyError(Exception):
    """
    Raised when an issues happens with respect to the store Key.
    """
    pass



from abc import ABC, abstractmethod
from datetime import datetime
from typing import List, Optional, Tuple, Any, Dict, Iterable, Generator

import yaml
from smart_open import smart_open

from blurr.core import logging
from blurr.core.aggregate_block import BlockAggregate, TimeAggregate
from blurr.core.errors import PrepareWindowMissingBlocksError
from blurr.core.evaluation import Context
from blurr.core.record import Record
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store import Store
from blurr.core.store_key import Key
from blurr.core.transformer_streaming import StreamingTransformer, StreamingTransformerSchema
from blurr.core.transformer_window import WindowTransformer
from blurr.core.type import Type
from blurr.runner.data_processor import DataProcessor

TimeAndRecord = Tuple[datetime, Record]


class Runner(ABC):
    """
    An abstract class that provides functionality to:
    - Convert raw events in Records
    - Process a list of Records for a user.

    A class that inherits from Runner should do the following:
    1. Call `get_per_identity_records()` using an iterator of the events available. This returns
        a generator which creates a Tuple[Identity, TimeAndRecord]] output.
    2. The Tuple[Identity, TimeAndRecord]] output should be grouped together by the
         Identity to create a List of TimeAndRecord per identity.
    3. Using the per identity list of TimeAndRecord `execute_per_identity_records()`
        should be called.
        - This returns Tuple[Identity, Tuple[Streaming BTS State, List of Window BTS output]].
        - `execute_per_identity_records()` can take in a existing old_state (old Streaming BTS
            State) so as to allow batch execution to make use of previous output.
    """

    def __init__(self, stream_bts_file: str, window_bts_file: Optional[str]):
        self._stream_bts = yaml.safe_load(smart_open(stream_bts_file))
        self._window_bts = None if window_bts_file is None else yaml.safe_load(
            smart_open(window_bts_file))

        # TODO: Assume validation will be done separately.
        # This causes a problem when running the code on spark
        # as the validation yml file is inside the archived package
        # and yamale is not able to read that.
        # validate_schema_spec(self._stream_bts)
        # if self._window_bts is not None:
        #     validate_schema_spec(self._window_bts)

    def execute_per_identity_records(
            self,
            identity: str,
            records: List[TimeAndRecord],
            old_state: Optional[Dict[Key, Any]] = None) -> Tuple[str, Tuple[Dict, List]]:
        """
        Executes the streaming and window BTS on the given records. An option old state can provided
        which initializes the state for execution. This is useful for batch execution where the
        previous state is written out to storage and can be loaded for the next batch run.

        :param identity: Identity of the records.
        :param records: List of TimeAndRecord to be processed.
        :param old_state: Streaming BTS state dictionary from a previous execution.
        :return: Tuple[Identity, Tuple[Identity, Tuple[Streaming BTS state dictionary,
            List of window BTS output]].
        """
        schema_loader = SchemaLoader()
        if records:
            records.sort(key=lambda x: x[0])
        else:
            records = []

        block_data = self._execute_stream_bts(records, identity, schema_loader, old_state)
        window_data = self._execute_window_bts(identity, schema_loader)

        return identity, (block_data, window_data)

    def get_per_identity_records(self, events: Iterable, data_processor: DataProcessor
                                 ) -> Generator[Tuple[str, TimeAndRecord], None, None]:
        """
        Uses the given iteratable events and the data processor convert the event into a list of
        Records along with its identity and time.
        :param events: iteratable events.
        :param data_processor: DataProcessor to process each event in events.
        :return: yields Tuple[Identity, TimeAndRecord] for all Records in events,
        """
        schema_loader = SchemaLoader()
        stream_bts_name = schema_loader.add_schema_spec(self._stream_bts)
        stream_transformer_schema: StreamingTransformerSchema = schema_loader.get_schema_object(
            stream_bts_name)
        for event in events:
            try:
                for record in data_processor.process_data(event):
                    try:
                        id = stream_transformer_schema.get_identity(record)
                        time = stream_transformer_schema.get_time(record)
                        yield (id, (time, record))
                    except Exception as err:
                        logging.error('{} in parsing Record {}.'.format(err, record))
            except Exception as err:
                logging.error('{} in parsing Event {}.'.format(err, event))

    def _execute_stream_bts(self,
                            identity_events: List[TimeAndRecord],
                            identity: str,
                            schema_loader: SchemaLoader,
                            old_state: Optional[Dict] = None) -> Dict[Key, Any]:
        if self._stream_bts is None:
            return {}

        stream_bts_name = schema_loader.add_schema_spec(self._stream_bts)
        stream_transformer_schema = schema_loader.get_schema_object(stream_bts_name)
        store = self._get_store(schema_loader)

        if old_state:
            for k, v in old_state.items():
                store.save(k, v)

        if identity_events:
            stream_transformer = StreamingTransformer(stream_transformer_schema, identity)

            for time, event in identity_events:
                stream_transformer.run_evaluate(event)
            stream_transformer.run_finalize()

        return self._get_store(schema_loader).get_all(identity)

    def _execute_window_bts(self, identity: str, schema_loader: SchemaLoader) -> List[Dict]:
        if self._window_bts is None:
            logging.debug('Window BTS not provided')
            return []

        stream_transformer = StreamingTransformer(
            self._get_streaming_transformer_schema(schema_loader), identity)
        all_data = self._get_store(schema_loader).get_all(identity)
        stream_transformer.run_restore(all_data)

        exec_context = Context()
        exec_context.add(stream_transformer._schema.name, stream_transformer)

        block_obj = None
        for aggregate in stream_transformer._nested_items.values():
            if not isinstance(aggregate, TimeAggregate):
                continue
            if block_obj is not None:
                raise Exception(('Window operation is supported against Streaming ',
                                 'BTS with only one BlockAggregate'))
            block_obj = aggregate

        if block_obj is None:
            raise Exception('No BlockAggregate found in the Streaming BTS file')

        window_data = []

        window_bts_name = schema_loader.add_schema_spec(self._window_bts)
        window_transformer_schema = schema_loader.get_schema_object(window_bts_name)
        window_transformer = WindowTransformer(window_transformer_schema, identity, exec_context)

        logging.debug('Running Window BTS for identity {}'.format(identity))

        anchors = 0
        blocks = 0
        for key, data in all_data.items():
            if key.group != block_obj._schema.name:
                continue
            try:
                blocks += 1
                if window_transformer.run_evaluate(block_obj.run_restore(data)):
                    anchors += 1
                    window_data.append(window_transformer.run_flattened_snapshot)
            except PrepareWindowMissingBlocksError as err:
                logging.debug('{} with {}'.format(err, key))

        if anchors == 0:
            logging.debug('No anchors found for identity {} out of {} blocks'.format(
                identity, blocks))

        return window_data

    @staticmethod
    def _get_store(schema_loader: SchemaLoader) -> Store:
        stores = schema_loader.get_all_stores()
        if not stores:
            fq_name_and_schema = schema_loader.get_schema_specs_of_type(
                Type.BLURR_STORE_DYNAMO, Type.BLURR_STORE_MEMORY)
            return schema_loader.get_store(next(iter(fq_name_and_schema)))

        return stores[0]

    @staticmethod
    def _get_streaming_transformer_schema(
            schema_loader: SchemaLoader) -> StreamingTransformerSchema:
        fq_name_and_schema = schema_loader.get_schema_specs_of_type(Type.BLURR_TRANSFORM_STREAMING)
        return schema_loader.get_schema_object(next(iter(fq_name_and_schema)))

    @abstractmethod
    def execute(self, *args, **kwargs):
        NotImplemented('execute must be implemented')

    @abstractmethod
    def write_output_file(self, *args, **kwargs):
        NotImplemented('execute must be implemented')

    @abstractmethod
    def print_output(self, *args, **kwargs):
        NotImplemented('execute must be implemented')



import os
import tempfile
import zipfile
from distutils.dir_util import copy_tree

from pip._internal import main as pip_main

from blurr.cli.util import eprint
from blurr.core import logging


def _create_zip(dir: str, target_file: str) -> None:
    dir_abs_path = os.path.abspath(dir)
    zip_file = zipfile.ZipFile(target_file, 'w', zipfile.ZIP_DEFLATED)
    for root, dirs, files in os.walk(dir):
        for file in files:
            file_path = os.path.join(root, file)
            zip_file.write(file_path, file_path.replace(dir_abs_path, ''))
    zip_file.close()


def _resolve_pip_dependencies(tmp_dir: str) -> int:
    pip_verbosity = [] if logging.is_debug_enabled() else ["--quiet"]
    pip_cmd = [
        "install", "--requirement",
        os.path.join(tmp_dir, "requirements.txt"), "--target", tmp_dir
    ] + pip_verbosity
    return pip_main(pip_cmd)


def package_spark(source_dir: str, target_file: str) -> int:
    if not os.path.isdir(source_dir):
        eprint(source_dir + " is not a valid directory")
        return 1

    req_file = os.path.join(source_dir, "requirements.txt")
    if not os.path.exists(req_file):
        eprint("requirements.txt not found in " + source_dir)
        return 1

    tmp_dir = tempfile.mkdtemp()

    logging.debug("copying app files to " + tmp_dir)
    copy_tree(source_dir, tmp_dir)

    logging.debug("resolving pip dependencies into " + tmp_dir)
    pip_out = _resolve_pip_dependencies(tmp_dir)
    if pip_out != 0:
        eprint("there was an error processing " + req_file)
        return 1

    logging.debug("creating package " + target_file)
    _create_zip(tmp_dir, target_file)
    print("spark app generated successfully: " + os.path.abspath(target_file))

    return 0



import os
import sys
from typing import List, Tuple, Dict

import yaml

from blurr.core.type import Type


def is_window_bts(bts_dict: Dict) -> bool:
    return Type.is_type_equal(bts_dict.get('Type', ''), Type.BLURR_TRANSFORM_WINDOW)


def is_streaming_bts(bts_dict: Dict) -> bool:
    return Type.is_type_equal(bts_dict.get('Type', ''), Type.BLURR_TRANSFORM_STREAMING)


def get_yml_files(path: str = '.') -> List[str]:
    return [
        os.path.join(path, file)
        for file in os.listdir(path)
        if (file.endswith('.yml') or file.endswith('.yaml'))
    ]


def get_stream_window_bts_files(bts_files: List[str]) -> Tuple[str, str]:
    stream_bts_file = None
    window_bts_file = None
    for bts_file in bts_files:
        if stream_bts_file and window_bts_file:
            break

        bts_dict = yaml.safe_load(open(bts_file))
        if not stream_bts_file and is_streaming_bts(bts_dict):
            stream_bts_file = bts_file

        if not window_bts_file and is_window_bts(bts_dict):
            window_bts_file = bts_file

    return stream_bts_file, window_bts_file


def eprint(*args):
    print(*args, file=sys.stderr)



import logging

import sys
import threading

# Don't use this directly. Use _get_logger() instead.
_logger = None
_logger_lock = threading.Lock()


def _get_logger() -> logging.Logger:
    global _logger

    # Use double-checked locking to avoid taking lock unnecessarily.
    if _logger:
        return _logger

    _logger_lock.acquire()

    try:
        if _logger:
            return _logger

        # Scope the blurr logger to not conflict with users' loggers.
        logger = logging.getLogger('blurr')

        # Don't further configure the blurr logger if the root logger is
        # already configured. This prevents double logging in those cases.
        if not logging.getLogger().handlers:
            # Determine whether we are in an interactive environment
            interactive = False
            try:
                # This is only defined in interactive shells.
                if sys.ps1:
                    interactive = True
            except AttributeError:
                # Even now, we may be in an interactive shell with `python -i`.
                interactive = sys.flags.interactive

            # If we are in an interactive environment (like Jupyter), set loglevel
            # to INFO and pipe the output to stdout.
            if interactive:
                logger.setLevel(logging.INFO)
                _logging_target = sys.stdout
            else:
                _logging_target = sys.stderr

            # Add the output handler.
            handler = logging.StreamHandler(_logging_target)
            handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT, None))
            logger.addHandler(handler)

        _logger = logger
        return _logger

    finally:
        _logger_lock.release()


def debug(msg, *args, **kwargs):
    _get_logger().debug(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    _get_logger().error(msg, *args, **kwargs)


def exception(msg, *args, **kwargs):
    _get_logger().exception(msg, *args, **kwargs)


def fatal(msg, *args, **kwargs):
    _get_logger().fatal(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    _get_logger().info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    _get_logger().warning(msg, *args, **kwargs)


def is_debug_enabled():
    _get_logger().isEnabledFor(logging.DEBUG)



from abc import abstractmethod, ABC
from datetime import datetime, timezone
from typing import Any, List, Tuple, Dict

from blurr.core.base import BaseSchema
from blurr.core.store_key import Key, KeyType


class StoreSchema(BaseSchema):
    pass


class Store(ABC):
    """ Base Store that allows for data to be persisted during / after transformation """

    @abstractmethod
    def get_all(self, identity: str) -> Dict[Key, Any]:
        """
        Gets all the items for an identity
        """
        raise NotImplementedError()

    @abstractmethod
    def get(self, key: Key) -> Any:
        """
        Gets an item by key
        """
        raise NotImplementedError()

    def get_range(self,
                  base_key: Key,
                  start_time: datetime,
                  end_time: datetime = None,
                  count: int = 0) -> List[Tuple[Key, Any]]:
        """
        Returns the list of items from the store based on the given time range or count.
        :param base_key: Items which don't start with the base_key are filtered out.
        :param start_time: Start time to for the range query
        :param end_time: End time of the range query. If None count is used.
        :param count: The number of items to be returned. Used if end_time is not specified.
        """
        if end_time and count:
            raise ValueError('Only one of `end` or `count` can be set')

        if count:
            end_time = datetime.min.replace(
                tzinfo=timezone.utc) if count < 0 else datetime.max.replace(tzinfo=timezone.utc)

        end_time = self._add_timezone_if_required(end_time)
        start_time = self._add_timezone_if_required(start_time)

        if end_time < start_time:
            start_time, end_time = end_time, start_time

        if base_key.key_type == KeyType.TIMESTAMP:
            start_key = Key(KeyType.TIMESTAMP, base_key.identity, base_key.group, [], start_time)
            end_key = Key(KeyType.TIMESTAMP, base_key.identity, base_key.group, [], end_time)
            return self._get_range_timestamp_key(start_key, end_key, count)
        else:
            return self._get_range_dimension_key(base_key, start_time, end_time, count)

    @abstractmethod
    def _get_range_timestamp_key(self, start: Key, end: Key,
                                 count: int = 0) -> List[Tuple[Key, Any]]:
        """
        Returns the list of items from the store based on the given time range or count.

        This is used when the key being used is a TIMESTAMP key.
        """
        raise NotImplementedError()

    def get_time_range(self, identity, group, start_time, end_time) -> List[Tuple[Key, Any]]:
        raise NotImplementedError()

    def get_count_range(self, identity, group, time, count):
        raise NotImplementedError()

    @abstractmethod
    def _get_range_dimension_key(self,
                                 base_key: Key,
                                 start_time: datetime,
                                 end_time: datetime,
                                 count: int = 0) -> List[Tuple[Key, Any]]:
        """
        Returns the list of items from the store based on the given time range or count.

        This is used when the key being used is a DIMENSION key.
        """
        raise NotImplementedError()

    @staticmethod
    def _restrict_items_to_count(items: List[Tuple[Key, Any]], count: int) -> List[Tuple[Key, Any]]:
        """
        Restricts items to count number if len(items) is larger than abs(count). This function
        assumes that items is sorted by time.

        :param items: The items to restrict.
        :param count: The number of items returned.
        """
        if abs(count) > len(items):
            count = Store._sign(count) * len(items)

        if count < 0:
            return items[count:]
        else:
            return items[:count]

    @abstractmethod
    def save(self, key: Key, item: Any) -> None:
        """
        Saves an item to store
        """
        raise NotImplementedError()

    @abstractmethod
    def delete(self, key: Key) -> None:
        """
        Deletes an item from the store by key
        """
        raise NotImplementedError()

    @abstractmethod
    def finalize(self) -> None:
        """
        Finalizes the store by flushing all remaining data to persistence
        """
        raise NotImplementedError()

    @staticmethod
    def _sign(x: int) -> int:
        return (1, -1)[x < 0]

    @staticmethod
    def _add_timezone_if_required(time: datetime) -> datetime:
        if time.tzinfo is None or time.tzinfo.utcoffset(time) is None:
            time = time.replace(tzinfo=timezone.utc)

        return time



from blurr.core.aggregate_identity import IdentityAggregateSchema, IdentityAggregate
from blurr.core.aggregate_time import TimeAggregateSchema, TimeAggregate


class BlockAggregateSchema(IdentityAggregateSchema, TimeAggregateSchema):
    """ Rolls up records into aggregate blocks.  Blocks are created when the split condition executes to true.  """

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_DIMENSIONS)


class BlockAggregate(IdentityAggregate, TimeAggregate):
    """
    Manages the aggregates for block based roll-ups of streaming data that are split by time.
    """
    pass



"""
Usage:
    blurr validate [--debug] [<BTS> ...]
    blurr transform [--debug] [--runner=<runner>] [--streaming-bts=<bts-file>] [--window-bts=<bts-file>] [--data-processor=<data-processor>] (--source=<raw-json-files> | <raw-json-files>)
    blurr package-spark [--debug] [--source-dir=<dir>] [--target=<zip-file>]
    blurr -h | --help

Commands:
    validate        Runs syntax validation on the list of BTS files provided. If
                    no files are provided then all *.bts files in the current
                    directory are validated.
    transform       Runs blurr to process the given raw log file. This command
                    can be run with the following combinations:
                    1. No BTS provided - The current directory is searched for
                    BTSs. First streaming and the first window BTS are used.
                    2. Only streaming BTS given - Transform outputs the result of
                    applying the BTS on the raw data file.
                    3. Both streaming and window BTS are provided - Transform
                    outputs the final result of applying the streaming and window
                    BTS on the raw data file.
    package-spark   Generates a submittable Spark app with .zip extension. Requires a
                    requirements.txt inside --source-dir. The generated package
                    will contain all the python code inside the provided --source-dir,
                    along with all the dependencies resolved from requirements.txt.


Options:
    -h --help                   Show this screen.
    --version                   Show version.
    --runner=<runner>           The runner to use for the transform.
                                Possible values:
                                local - Transforms done in memory. <default>
                                spark - Transforms done using spark locally.
    --streaming-bts=<bts-file>  Streaming BTS file to use.
    --window-bts=<bts-file>     Window BTS file to use.
    --source=<raw-json-files>   List of source files separated by comma
    --debug                     Output debug logs.
    --data-processor=<data-processor>   Data processor to use to process each record.
                                        Possible values:
                                        simple - One event dictionary per line in the source file(s). <default>
                                        ipfix - Processor for IpFix format.
    --source-dir=<dir>          A directory containing a Spark app to be packaged. [default: ./]
    --target=<zip-file>         Filename of the generated Spark app zipfile. [default: spark-app.zip]
"""
import logging
import os
import sys

from docopt import docopt

from blurr.cli.cli import cli

PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))
VERSION_PATH = os.path.join(PACKAGE_DIR, 'VERSION')


def read_version(version_file: str) -> str:
    if os.path.exists(version_file) and os.path.isfile(version_file):
        with open(version_file, 'r') as file:
            version = file.readline()
            file.close()
        return version
    else:
        return 'LOCAL'


def setup_logs(debug: bool) -> None:
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    _handler = logging.StreamHandler(sys.stdout)
    _handler.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))
    logger.addHandler(_handler)


def main():
    arguments = docopt(__doc__, version=read_version(VERSION_PATH))
    setup_logs(arguments['--debug'])
    result = cli(arguments)
    sys.exit(result)


if __name__ == '__main__':
    main()



from datetime import timedelta, datetime, timezone

from blurr.core.aggregate_time import TimeAggregateSchema, TimeAggregate
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store_key import Key, KeyType


class ActivityAggregateSchema(TimeAggregateSchema):
    """ Aggregates activity into blocks by partitioning around periods of inactivity """

    ATTRIBUTE_SEPARATE_BY_INACTIVE_SECONDS = 'SeparateByInactiveSeconds'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        inactive_seconds = self._spec.get(self.ATTRIBUTE_SEPARATE_BY_INACTIVE_SECONDS, 0)
        self.separation_interval: timedelta = timedelta(
            seconds=int(inactive_seconds) if str(inactive_seconds).isdigit() else 0)
        self.key_type = KeyType.TIMESTAMP

    def validate_schema_spec(self) -> None:
        super(TimeAggregateSchema, self).validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_SEPARATE_BY_INACTIVE_SECONDS)
        self.validate_number_attribute(self.ATTRIBUTE_SEPARATE_BY_INACTIVE_SECONDS, int, 1)


class ActivityAggregate(TimeAggregate):
    """ Aggregates activity in blocks separated by periods of inactivity"""

    def run_evaluate(self) -> None:
        time = self._evaluation_context.global_context['time']

        if not self._start_time:
            self._load_old_state()

        # If the event time is beyond separation threshold, create a new block
        if self._start_time and (time < self._start_time - self._schema.separation_interval
                                 or time > self._end_time + self._schema.separation_interval):
            self._persist()
            self.run_reset()

        super().run_evaluate()

    def _load_old_state(self):
        most_recent_block = self._store.get_range(
            Key(self._schema.key_type, self._identity, self._name), datetime.now(timezone.utc),
            None, -1)
        if most_recent_block:
            self.run_restore(most_recent_block[0][1])

    @property
    def _key(self):
        return Key(self._schema.key_type, self._identity, self._name, [], self._start_time)



from enum import Enum
from typing import Union, List


class Type(Enum):
    BLURR_TRANSFORM_STREAMING = "blurr:transform:streaming"
    BLURR_TRANSFORM_WINDOW = "blurr:transform:window"
    BLURR_AGGREGATE_BLOCK = "blurr:aggregate:block"
    BLURR_AGGREGATE_LABEL = "blurr:aggregate:label"
    BLURR_AGGREGATE_ACTIVITY = "blurr:aggregate:activity"
    BLURR_AGGREGATE_IDENTITY = "blurr:aggregate:identity"
    BLURR_AGGREGATE_VARIABLE = "blurr:aggregate:variable"
    BLURR_AGGREGATE_WINDOW = "blurr:aggregate:window"
    BLURR_STORE_MEMORY = "blurr:store:memory"
    BLURR_STORE_DYNAMO = "blurr:store:dynamo"
    ANCHOR = "anchor"
    DAY = "day"
    HOUR = "hour"
    COUNT = "count"
    STRING = "string"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    FLOAT = "float"
    MAP = "map"
    LIST = "list"
    SET = "set"

    @staticmethod
    def is_type_equal(actual_type: Union[str, 'Type'], expected_type: 'Type') -> bool:
        try:
            return Type(actual_type) == expected_type
        except ValueError:
            return False

    @staticmethod
    def is_type_in(actual_type: Union[str, 'Type'], expected_type: List['Type']) -> bool:
        try:
            return Type(actual_type) in expected_type
        except ValueError:
            return False

    @staticmethod
    def is_store_type(store_type: Union[str, 'Type']) -> bool:
        return Type.is_type_in(store_type, [Type.BLURR_STORE_MEMORY, Type.BLURR_STORE_DYNAMO])

    @staticmethod
    def contains(value: Union[str, 'Type']) -> bool:
        """ Checks if a type is defined """
        if isinstance(value, str):
            return any(value.lower() == i.value for i in Type)

        return any(value == i for i in Type)


# Override __new__ to handle intialization by value of any case.
Type.__new__ = lambda cls, value: (super(Type, cls).__new__(cls, value.lower())
                                   if isinstance(value, str) else
                                   super(Type, cls).__new__(cls, value))



from abc import ABC
from typing import List, Dict, Any

from blurr.core.aggregate import AggregateSchema, Aggregate
from blurr.core.type import Type
from blurr.core.validator import ATTRIBUTE_INTERNAL


class TimeAggregateSchema(AggregateSchema, ABC):
    """ Rolls up records into aggregate blocks.  Blocks are created when the split condition executes to true.  """

    def extend_schema_spec(self) -> None:
        """ Injects the block start and end times """
        super().extend_schema_spec()

        if self.ATTRIBUTE_FIELDS in self._spec:
            # Add new fields to the schema spec. Since `_identity` is added by the super, new elements are added after
            predefined_field = self._build_time_fields_spec(self._spec[self.ATTRIBUTE_NAME])
            self._spec[self.ATTRIBUTE_FIELDS][1:1] = predefined_field

            # Add new field schema to the schema loader
            for field_schema in predefined_field:
                self.schema_loader.add_schema_spec(field_schema, self.fully_qualified_name)

    @staticmethod
    def _build_time_fields_spec(name_in_context: str) -> List[Dict[str, Any]]:
        """
        Constructs the spec for predefined fields that are to be included in the master spec prior to schema load
        :param name_in_context: Name of the current object in the context
        :return:
        """
        return [
            {
                'Name': '_start_time',
                'Type': Type.DATETIME,
                'Value': ('time if {aggregate}._start_time is None else time '
                          'if time < {aggregate}._start_time else {aggregate}._start_time'
                          ).format(aggregate=name_in_context),
                ATTRIBUTE_INTERNAL: True
            },
            {
                'Name': '_end_time',
                'Type': Type.DATETIME,
                'Value': ('time if {aggregate}._end_time is None else time '
                          'if time > {aggregate}._end_time else {aggregate}._end_time'
                          ).format(aggregate=name_in_context),
                ATTRIBUTE_INTERNAL: True
            },
        ]


class TimeAggregate(Aggregate, ABC):
    pass



from datetime import datetime, timezone

from blurr.core.errors import IdentityError, TimeError
from blurr.core.field_simple import DateTimeFieldSchema
from blurr.core.record import Record
from blurr.core.schema_loader import SchemaLoader
from blurr.core.transformer import Transformer, TransformerSchema


class StreamingTransformerSchema(TransformerSchema):
    """
    Represents the schema for processing streaming data.  Handles the streaming specific attributes of the schema
    """

    ATTRIBUTE_IDENTITY = 'Identity'
    ATTRIBUTE_TIME = 'Time'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.identity = self.build_expression(self.ATTRIBUTE_IDENTITY)
        self.time = self.build_expression(self.ATTRIBUTE_TIME)

    def validate_schema_spec(self) -> None:
        super().validate_schema_spec()
        self.validate_required_attributes(self.ATTRIBUTE_IDENTITY, self.ATTRIBUTE_TIME,
                                          self.ATTRIBUTE_STORES)

    def get_identity(self, record: Record) -> str:
        """
        Evaluates and returns the identity as specified in the schema.
        :param record: Record which is used to determine the identity.
        :return: The evaluated identity
        :raises: IdentityError if identity cannot be determined.
        """
        context = self.schema_context.context
        context.add_record(record)
        identity = self.identity.evaluate(context)
        if not identity:
            raise IdentityError('Could not determine identity using {}. Record is {}'.format(
                self.identity.code_string, record))
        context.remove_record()
        return identity

    def get_time(self, record: Record) -> datetime:
        context = self.schema_context.context
        context.add_record(record)
        time = self.time.evaluate(context)

        if not time or not isinstance(time, datetime):
            raise TimeError('Could not determine time using {}.  Record is {}'.format(
                self.time.code_string, record))

        time = DateTimeFieldSchema.sanitize_object(time)

        context.remove_record()
        return time


class StreamingTransformer(Transformer):
    def __init__(self, schema: TransformerSchema, identity: str) -> None:
        super().__init__(schema, identity)
        self._evaluation_context.global_add('identity', self._identity)

    def run_evaluate(self, record: Record):
        """
        Evaluates and updates data in the StreamingTransformer.
        :param record: The 'source' record used for the update.
        :raises: IdentityError if identity is different from the one used during
        initialization.
        """
        record_identity = self._schema.get_identity(record)
        if self._identity != record_identity:
            raise IdentityError(
                'Identity in transformer ({}) and new record ({}) do not match'.format(
                    self._identity, record_identity))

        # Add source record and time to the global context
        self._evaluation_context.add_record(record)
        self._evaluation_context.global_add(
            'time',
            DateTimeFieldSchema.sanitize_object(
                self._schema.time.evaluate(self._evaluation_context)))
        super().run_evaluate()

        # Cleanup source and time form the context
        self._evaluation_context.remove_record()
        self._evaluation_context.global_remove('time')



import os

from pipenv.project import Project
from pipenv.utils import convert_deps_to_pip
from setuptools import setup, find_packages
'''
This setup.py script is expected to perform development and final releases using 
CircleCI (https://circleci.com/gh/productml/blurr).

 - Release Builds are created from tags, using CIRCLE_TAG environment variable
 - Development builds create a different artifact, 'blurr-dev', that can be installed
   using 'pip install blurr-dev'
'''


def is_ci_build():
    return True if os.getenv('CIRCLECI') else False


def is_release():
    if not is_ci_build():
        return False
    return True if os.getenv('CIRCLE_TAG') else False


def requirements():
    pipfile = Project(chdir=False).parsed_pipfile
    return convert_deps_to_pip(pipfile['packages'], r=False)


def readme():
    with open('README.md', 'r', encoding='utf-8') as f:
        return f.read()


def name():
    return "blurr" if is_release() else "blurr-dev"


def version():
    if not is_ci_build():
        return "LOCAL"
    elif is_release():
        return os.getenv('CIRCLE_TAG')
    else:
        return "0." + os.getenv('CIRCLE_BUILD_NUM')


# emit a VERSION file the CLI can use to check current version
version_file = open("blurr/VERSION", "w")
version_file.write(version())
version_file.close()

setup(
    name=name(),
    version=version(),
    description="Data aggregation pipeline for running real-time predictive models",
    long_description=readme(),
    long_description_content_type='text/markdown',
    author="productml.com",
    author_email="info@productml.com",
    url="https://github.com/productml/blurr",
    packages=find_packages(),
    data_files=["blurr/VERSION"],
    include_package_data=True,
    install_requires=requirements(),
    python_requires='>=3.6',
    classifiers=[
        "Development Status :: 1 - Planning",  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
    ],
    entry_points={'console_scripts': ['blurr = blurr.__main__:main']})



from copy import copy
from enum import Enum
from typing import Any, Dict

from blurr.core import logging
from blurr.core.errors import MissingAttributeError
from blurr.core.record import Record


class Context(dict):
    """
    Evaluation context provides a dictionary of declared context objects
    """

    def __init__(self, initial_value: Dict[str, Any] = None):
        """
        Initializes a new context with an existing dictionary
        :param initial_value: Dictionary mapping strings to execution context objects
        """
        if initial_value:
            super().__init__(initial_value)

    def add(self, name, value):
        """
        Adds a context item by name
        :param name: Name of the item in the context for evaluation
        :param value: Object that the name refers to in the context
        """
        self[name] = value

    def merge(self, context: 'Context') -> None:
        """
        Updates the context object with the items in another context
        object.  Fields of the same name are overwritten
        :param context: Another context object to superimpose on current
        :return:
        """
        self.update(context)


class EvaluationContext:
    def __init__(self, global_context: Context = None, local_context: Context = None) -> None:
        """
        Initializes an evaluation context with global and local context dictionaries.  The unset parameters default
        to an empty context dictionary.
        :param global_context: Global context dictionary
        :param local_context: Local context dictionary.
        """
        self.global_context = Context() if global_context is None else global_context
        self.local_context = Context() if local_context is None else local_context

    @property
    def fork(self) -> 'EvaluationContext':
        """
        Creates a copy of the current evaluation config with the same global context, but a shallow copy of the
        local context
        :return:
        """
        return EvaluationContext(self.global_context, copy(self.local_context))

    def local_include(self, context: Context) -> None:
        """
        Includes the supplied context into the local context
        :param context: Context to include into local context
        """
        self.local_context.merge(context)

    def global_add(self, key: str, value: Any) -> None:
        """
        Adds a key and value to the global dictionary
        """
        self.global_context[key] = value

    def global_remove(self, key: str) -> None:
        """
        Removes the key and its value from the global dictionary
        """
        del self.global_context[key]

    def add_record(self, record: Record) -> None:
        """
        Adds a record to the evaluation context.
        :param record: Record to be added
        """
        self.global_add('source', record)

    def remove_record(self) -> None:
        """Removes any previously added record to the evaluation context."""
        del self.global_context['source']

    def merge(self, evaluation_context: 'EvaluationContext') -> None:
        """
        Merges the provided evaluation context to the current evaluation context.
        :param evaluation_context: Evaluation context to merge.
        """
        self.global_context.merge(evaluation_context.global_context)
        self.local_context.merge(evaluation_context.local_context)


class ExpressionType(Enum):
    EVAL = 'eval'
    EXEC = 'exec'


class Expression:
    """ Encapsulates a python code statement in string and in compilable expression"""

    def __init__(self, code_string: str,
                 expression_type: ExpressionType = ExpressionType.EVAL) -> None:
        """
        An expression must be initialized with a python statement
        :param code_string: Python code statement
        """

        self.code_string = str(code_string)
        self.type = expression_type

        self.code_object = compile(self.code_string, '<string>', self.type.value)

    def evaluate(self, evaluation_context: EvaluationContext) -> Any:
        """
        Evaluates the expression with the context provided.  If the execution
        results in failure, an ExpressionEvaluationException encapsulating the
        underlying exception is raised.
        :param evaluation_context: Global and local context dictionary to be passed for evaluation
        """
        try:

            if self.type == ExpressionType.EVAL:
                return eval(self.code_object, evaluation_context.global_context,
                            evaluation_context.local_context)

            elif self.type == ExpressionType.EXEC:
                return exec(self.code_object, evaluation_context.global_context,
                            evaluation_context.local_context)

        except Exception as err:
            # Evaluation exceptions are expected because of missing fields in the source 'Record'.
            logging.debug('{} in evaluating expression {}. Error: {}'.format(
                type(err).__name__, self.code_string, err))
            # These should result in an exception being raised:
            # NameError - Exceptions thrown because of using names in the expression which are not
            #   present in EvaluationContext. A common cause for this is typos in the BTS.
            # MissingAttributeError - Exception thrown when a BTS nested item is used which does not
            #   exist. Should only happen for erroneous BTSs.
            # ImportError - Thrown when there is a failure in importing other modules.
            if isinstance(err, (NameError, MissingAttributeError, ImportError)):
                raise err
            return None



from typing import Dict, Any

from blurr.cli.package_spark import package_spark
from blurr.cli.transform import transform
from blurr.cli.validate import validate_command


def cli(arguments: Dict[str, Any]) -> int:
    if arguments['validate']:
        return validate_command(arguments['<BTS>'])
    elif arguments['transform']:
        source = []
        if arguments['--source'] is not None:
            source = arguments['--source'].split(',')
        elif arguments['<raw-json-files>'] is not None:
            source = arguments['<raw-json-files>'].split(',')
        return transform(arguments['--runner'], arguments['--streaming-bts'],
                         arguments['--window-bts'], arguments['--data-processor'], source)
    elif arguments['package-spark']:
        return package_spark(arguments['--source-dir'], arguments['--target'])



from typing import Dict, List

from blurr.core.evaluation import Expression, EvaluationContext, ExpressionType


class SchemaContext:

    ATTRIBUTE_IMPORT_MODULE = 'Module'
    ATTRIBUTE_IMPORT_IDENTIFIERS = 'Identifiers'

    def __init__(self, import_spec: List[Dict]):
        self.import_spec = import_spec
        self.import_statements = self._generate_import_statements()
        self._context = None

    def _generate_import_statements(self) -> List[Expression]:
        import_expression_list = []
        if self.import_spec is None:
            return import_expression_list

        for custom_import in self.import_spec:
            module = custom_import[self.ATTRIBUTE_IMPORT_MODULE]
            identifier_list = custom_import.get(self.ATTRIBUTE_IMPORT_IDENTIFIERS, None)

            statement = ' '.join(['import', module]) if not identifier_list else ' '.join(
                ['from', module, 'import', ','.join(identifier_list)])
            import_expression_list.append(Expression(statement, ExpressionType.EXEC))

        return import_expression_list

    @property
    def context(self) -> EvaluationContext:
        if self._context:
            return self._context
        # The eval code adds the python global context to the global context dict being passed and
        # new context being created is added to the local context. We take the local_context in
        # temp_eval_context and use that as the global context for the returned EvaluationContext.
        temp_eval_context = EvaluationContext()
        for import_statement in self.import_statements:
            import_statement.evaluate(temp_eval_context)

        eval_context = EvaluationContext(global_context=temp_eval_context.local_context)

        return eval_context



from datetime import datetime, timezone
from enum import Enum
from typing import List, Any

from dateutil import parser


class KeyType(Enum):
    DIMENSION = 1
    TIMESTAMP = 2
    UNKNOWN = 10


class Key:
    """
    A record in the store is identified by a key
    """
    PARTITION = '/'
    DIMENSION_PARTITION = ':'

    # TODO: Consider adding a * to force parameterization of attributes.
    def __init__(self,
                 key_type: KeyType,
                 identity: str,
                 group: str,
                 dimensions: List[str] = list(),
                 timestamp: datetime = None) -> None:
        """
        Initializes a new key for storing data
        :param identity: Primary identity of the record being stored
        :param group: Secondary identity of the record
        :param timestamp: Optional timestamp that can be used for time range queries
        """
        if not identity or identity.isspace():
            raise ValueError('`identity` must be present.')

        if not group or group.isspace():
            raise ValueError('`group` must be present.')

        if dimensions and timestamp:
            raise ValueError('Both dimensions and timestamp should not be set together.')

        if key_type == KeyType.DIMENSION and timestamp:
            raise ValueError('`timestamp` should not be set for KeyType.DIMENSION.')

        if key_type == KeyType.TIMESTAMP and dimensions:
            raise ValueError('`dimensions` should not be set for KeyType.TIMESTAMP.')

        self.key_type = key_type
        self.identity = identity
        self.group = group
        self.timestamp = timestamp if not timestamp or timestamp.tzinfo else timestamp.replace(
            tzinfo=timezone.utc)
        self.dimensions = dimensions

    # TODO: Handle '/' and ':' values in dimensions
    @property
    def dimensions_str(self):
        return ':'.join(self.dimensions) if self.dimensions else ''

    @staticmethod
    def parse(key_string: str) -> 'Key':
        """ Parses a flat key string and returns a key """
        parts = key_string.split(Key.PARTITION)
        key_type = KeyType.DIMENSION
        if parts[3]:
            key_type = KeyType.TIMESTAMP
        return Key(key_type, parts[0], parts[1], parts[2].split(Key.DIMENSION_PARTITION)
                   if parts[2] else [],
                   parser.parse(parts[3]) if parts[3] else None)

    @staticmethod
    def parse_sort_key(identity: str, sort_key_string: str) -> 'Key':
        """ Parses a flat key string and returns a key """
        parts = sort_key_string.split(Key.PARTITION)
        key_type = KeyType.DIMENSION
        if parts[2]:
            key_type = KeyType.TIMESTAMP
        return Key(key_type, identity, parts[0], parts[1].split(Key.DIMENSION_PARTITION)
                   if parts[1] else [],
                   parser.parse(parts[2]) if parts[2] else None)

    def __str__(self):
        """ Returns the string representation of the key"""
        return Key.PARTITION.join([self.identity, self.sort_key])

    @property
    def sort_key(self):
        return Key.PARTITION.join(
            [self.group, self.dimensions_str,
             self.timestamp.isoformat() if self.timestamp else ''])

    @property
    def sort_prefix_key(self):
        if self.key_type == KeyType.DIMENSION:
            return Key.PARTITION.join([self.group, self.dimensions_str]
                                      if self.dimensions_str else [self.group, ''])

        if self.key_type == KeyType.TIMESTAMP:
            return self.sort_key

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other: 'Key') -> bool:
        return other and (self.identity, self.group, self.timestamp,
                          self.dimensions) == (other.identity, other.group, other.timestamp,
                                               other.dimensions)

    def __lt__(self, other: 'Key') -> bool:
        """
        Does a less than comparison on two keys. A None timestamp is considered
        larger than a timestamp that has been set.
        """
        if (self.identity, self.group, self.key_type) != (other.identity, other.group,
                                                          other.key_type):
            return False

        if self.key_type == KeyType.TIMESTAMP:
            return self.timestamp < other.timestamp

        return self.dimensions < other.dimensions

    def __gt__(self, other: 'Key') -> bool:
        """
        Does a greater than comparison on two keys. A None timestamp is
        considered larger than a timestamp that has been set.
        """
        if (self.identity, self.group, self.key_type) != (other.identity, other.group,
                                                          other.key_type):
            return False

        if self.key_type == KeyType.TIMESTAMP:
            return self.timestamp > other.timestamp

        return self.dimensions > other.dimensions

    def __hash__(self):
        return hash((self.identity, self.group, self.timestamp, self.dimensions_str))

    def starts_with(self, other: 'Key') -> bool:
        """
        Checks if this key starts with the other key provided. Returns False if key_type, identity
        or group are different.
        For `KeyType.TIMESTAMP` returns True.
        For `KeyType.DIMENSION` does prefix match between the two dimensions property.
        """
        if (self.key_type, self.identity, self.group) != (other.key_type, other.identity,
                                                          other.group):
            return False
        if self.key_type == KeyType.TIMESTAMP:
            return True
        if self.key_type == KeyType.DIMENSION:
            if len(self.dimensions) < len(other.dimensions):
                return False
            return self.dimensions[0:len(other.dimensions)] == other.dimensions



from typing import Any


def wrap(value: Any) -> Any:
    if isinstance(value, dict):
        return Record(value)
    # Lists are returned as Record lists
    elif isinstance(value, list) and len(value) > 0:
        return RecordList(value)
    # Simple values are returned as-is
    return value


class Record(dict):
    """
    Wraps a dictionary into an object to allow dictionary keys to be accessed as object properties
    """

    def __getattr__(self, name):
        """
        When attributes are not found, None is returned
        """
        if name.startswith('__') and name.endswith('__'):
            raise AttributeError('Record object has no {} attribute.'.format(name))
        return wrap(self[name]) if name in self else None

    def __getitem__(self, item):
        return wrap(super().__getitem__(item)) if item in self else None


class RecordList(list):
    """ Wraps a list to list of Records"""

    def __getitem__(self, item):
        return wrap(super().__getitem__(item))
